from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class StubOracleCreateArgs(typing.TypedDict):
    price: types.i80f48.I80F48


layout = borsh.CStruct("price" / types.i80f48.I80F48.layout)


class StubOracleCreateAccounts(typing.TypedDict):
    group: PublicKey
    oracle: PublicKey
    admin: PublicKey
    mint: PublicKey
    payer: PublicKey


def stub_oracle_create(
    args: StubOracleCreateArgs,
    accounts: StubOracleCreateAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xac?eS\x8dL\xc7\xd8"
    encoded_args = layout.build(
        {
            "price": args["price"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PerpEquityJSON(typing.TypedDict):
    perp_market_index: int
    value: i80f48.I80F48JSON


@dataclass
class PerpEquity:
    layout: typing.ClassVar = borsh.CStruct(
        "perp_market_index" / borsh.U16, "value" / i80f48.I80F48.layout
    )
    perp_market_index: int
    value: i80f48.I80F48

    @classmethod
    def from_decoded(cls, obj: Container) -> "PerpEquity":
        return cls(
            perp_market_index=obj.perp_market_index,
            value=i80f48.I80F48.from_decoded(obj.value),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "perp_market_index": self.perp_market_index,
            "value": self.value.to_encodable(),
        }

    def to_json(self) -> PerpEquityJSON:
        return {
            "perp_market_index": self.perp_market_index,
            "value": self.value.to_json(),
        }

    @classmethod
    def from_json(cls, obj: PerpEquityJSON) -> "PerpEquity":
        return cls(
            perp_market_index=obj["perp_market_index"],
            value=i80f48.I80F48.from_json(obj["value"]),
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class ComputeAccountDataAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey


def compute_account_data(
    accounts: ComputeAccountDataAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"6F\xbeb\xf6\xf2ct"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class PerpCancelOrderArgs(typing.TypedDict):
    order_id: int


layout = borsh.CStruct("order_id" / borsh.U128)


class PerpCancelOrderAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey


def perp_cancel_order(
    args: PerpCancelOrderArgs,
    accounts: PerpCancelOrderAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xe9\t\xbdD\xe0\xa3\xf5\xc1"
    encoded_args = layout.build(
        {
            "order_id": args["order_id"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from ..program_id import PROGRAM_ID
from .. import types


class EventQueueJSON(typing.TypedDict):
    header: types.event_queue_header.EventQueueHeaderJSON
    buf: list[types.any_event.AnyEventJSON]
    reserved: list[int]


@dataclass
class EventQueue:
    discriminator: typing.ClassVar = b")\xd0t\xd1\xadt\x8dD"
    layout: typing.ClassVar = borsh.CStruct(
        "header" / types.event_queue_header.EventQueueHeader.layout,
        "buf" / types.any_event.AnyEvent.layout[488],
        "reserved" / borsh.U8[64],
    )
    header: types.event_queue_header.EventQueueHeader
    buf: list[types.any_event.AnyEvent]
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["EventQueue"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["EventQueue"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["EventQueue"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "EventQueue":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = EventQueue.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            header=types.event_queue_header.EventQueueHeader.from_decoded(dec.header),
            buf=list(
                map(lambda item: types.any_event.AnyEvent.from_decoded(item), dec.buf)
            ),
            reserved=dec.reserved,
        )

    def to_json(self) -> EventQueueJSON:
        return {
            "header": self.header.to_json(),
            "buf": list(map(lambda item: item.to_json(), self.buf)),
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: EventQueueJSON) -> "EventQueue":
        return cls(
            header=types.event_queue_header.EventQueueHeader.from_json(obj["header"]),
            buf=list(
                map(lambda item: types.any_event.AnyEvent.from_json(item), obj["buf"])
            ),
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class BidsJSON(typing.TypedDict):
    kind: typing.Literal["Bids"]


class AsksJSON(typing.TypedDict):
    kind: typing.Literal["Asks"]


@dataclass
class Bids:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Bids"

    @classmethod
    def to_json(cls) -> BidsJSON:
        return BidsJSON(
            kind="Bids",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Bids": {},
        }


@dataclass
class Asks:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Asks"

    @classmethod
    def to_json(cls) -> AsksJSON:
        return AsksJSON(
            kind="Asks",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Asks": {},
        }


OrderTreeTypeKind = typing.Union[Bids, Asks]
OrderTreeTypeJSON = typing.Union[BidsJSON, AsksJSON]


def from_decoded(obj: dict) -> OrderTreeTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Bids" in obj:
        return Bids()
    if "Asks" in obj:
        return Asks()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderTreeTypeJSON) -> OrderTreeTypeKind:
    if obj["kind"] == "Bids":
        return Bids()
    if obj["kind"] == "Asks":
        return Asks()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Bids" / borsh.CStruct(), "Asks" / borsh.CStruct())



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solana.publickey import PublicKey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class LeafNodeJSON(typing.TypedDict):
    tag: int
    owner_slot: int
    order_type: int
    padding: list[int]
    time_in_force: int
    padding2: list[int]
    key: int
    owner: str
    quantity: int
    timestamp: int
    peg_limit: int
    client_order_id: int
    reserved: list[int]


@dataclass
class LeafNode:
    layout: typing.ClassVar = borsh.CStruct(
        "tag" / borsh.U8,
        "owner_slot" / borsh.U8,
        "order_type" / borsh.U8,
        "padding" / borsh.U8[1],
        "time_in_force" / borsh.U16,
        "padding2" / borsh.U8[2],
        "key" / borsh.U128,
        "owner" / BorshPubkey,
        "quantity" / borsh.I64,
        "timestamp" / borsh.U64,
        "peg_limit" / borsh.I64,
        "client_order_id" / borsh.U64,
        "reserved" / borsh.U8[32],
    )
    tag: int
    owner_slot: int
    order_type: int
    padding: list[int]
    time_in_force: int
    padding2: list[int]
    key: int
    owner: PublicKey
    quantity: int
    timestamp: int
    peg_limit: int
    client_order_id: int
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "LeafNode":
        return cls(
            tag=obj.tag,
            owner_slot=obj.owner_slot,
            order_type=obj.order_type,
            padding=obj.padding,
            time_in_force=obj.time_in_force,
            padding2=obj.padding2,
            key=obj.key,
            owner=obj.owner,
            quantity=obj.quantity,
            timestamp=obj.timestamp,
            peg_limit=obj.peg_limit,
            client_order_id=obj.client_order_id,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "tag": self.tag,
            "owner_slot": self.owner_slot,
            "order_type": self.order_type,
            "padding": self.padding,
            "time_in_force": self.time_in_force,
            "padding2": self.padding2,
            "key": self.key,
            "owner": self.owner,
            "quantity": self.quantity,
            "timestamp": self.timestamp,
            "peg_limit": self.peg_limit,
            "client_order_id": self.client_order_id,
            "reserved": self.reserved,
        }

    def to_json(self) -> LeafNodeJSON:
        return {
            "tag": self.tag,
            "owner_slot": self.owner_slot,
            "order_type": self.order_type,
            "padding": self.padding,
            "time_in_force": self.time_in_force,
            "padding2": self.padding2,
            "key": self.key,
            "owner": str(self.owner),
            "quantity": self.quantity,
            "timestamp": self.timestamp,
            "peg_limit": self.peg_limit,
            "client_order_id": self.client_order_id,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: LeafNodeJSON) -> "LeafNode":
        return cls(
            tag=obj["tag"],
            owner_slot=obj["owner_slot"],
            order_type=obj["order_type"],
            padding=obj["padding"],
            time_in_force=obj["time_in_force"],
            padding2=obj["padding2"],
            key=obj["key"],
            owner=PublicKey(obj["owner"]),
            quantity=obj["quantity"],
            timestamp=obj["timestamp"],
            peg_limit=obj["peg_limit"],
            client_order_id=obj["client_order_id"],
            reserved=obj["reserved"],
        )



import argparse
import asyncio

from ..mango_client import MangoClient


async def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--mango-account',
        help='Mango account primary key.',
        required=True
    )

    args = parser.parse_args()

    mango_client = await MangoClient.connect()

    mango_account = await mango_client.get_mango_account(args.mango_account)

    print(await mango_client.health_ratio(mango_account))


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class PerpConsumeEventsArgs(typing.TypedDict):
    limit: int


layout = borsh.CStruct("limit" / borsh.U64)


class PerpConsumeEventsAccounts(typing.TypedDict):
    group: PublicKey
    perp_market: PublicKey
    event_queue: PublicKey


def perp_consume_events(
    args: PerpConsumeEventsArgs,
    accounts: PerpConsumeEventsAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x9eU\x1d\xd18\xeb %"
    encoded_args = layout.build(
        {
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class AnyEventJSON(typing.TypedDict):
    event_type: int
    padding: list[int]


@dataclass
class AnyEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "event_type" / borsh.U8, "padding" / borsh.U8[207]
    )
    event_type: int
    padding: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "AnyEvent":
        return cls(event_type=obj.event_type, padding=obj.padding)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"event_type": self.event_type, "padding": self.padding}

    def to_json(self) -> AnyEventJSON:
        return {"event_type": self.event_type, "padding": self.padding}

    @classmethod
    def from_json(cls, obj: AnyEventJSON) -> "AnyEvent":
        return cls(event_type=obj["event_type"], padding=obj["padding"])



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class AccountExpandArgs(typing.TypedDict):
    token_count: int
    serum3_count: int
    perp_count: int
    perp_oo_count: int


layout = borsh.CStruct(
    "token_count" / borsh.U8,
    "serum3_count" / borsh.U8,
    "perp_count" / borsh.U8,
    "perp_oo_count" / borsh.U8,
)


class AccountExpandAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    payer: PublicKey


def account_expand(
    args: AccountExpandArgs,
    accounts: AccountExpandAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"X\xd4\x1ft\xfd\xc9Q\x01"
    encoded_args = layout.build(
        {
            "token_count": args["token_count"],
            "serum3_count": args["serum3_count"],
            "perp_count": args["perp_count"],
            "perp_oo_count": args["perp_oo_count"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from mango_explorer_v4.accounts.perp_market import PerpMarket
from mango_explorer_v4.constants import QUOTE_DECIMALS
from decimal import Decimal


class PerpMarketHelper():
    @staticmethod
    async def oracle_price(perp_market) -> Decimal:
        pass

    @staticmethod
    def price_lots_to_ui(perp_market: PerpMarket, price_lots: int):
        return float(Decimal(price_lots) * Decimal(10) ** Decimal(perp_market.base_decimals - QUOTE_DECIMALS) * Decimal(perp_market.quote_lot_size) / Decimal(perp_market.base_lot_size))

    @staticmethod
    def base_lots_to_ui(perp_market: PerpMarket, base_lots: int):
        return float(Decimal(base_lots) * (Decimal(perp_market.base_lot_size) / Decimal(10) ** Decimal(perp_market.base_decimals)))

    @staticmethod
    def quote_lots_to_ui(perp_market: PerpMarket, quote_lots: int):
        raise NotImplementedError

    @staticmethod
    def ui_price_to_lots(perp_market: PerpMarket, ui_price: float) -> int:
        return int(Decimal(ui_price * 10 ** QUOTE_DECIMALS) * Decimal(perp_market.base_lot_size) / Decimal(
            perp_market.quote_lot_size * 10 ** perp_market.base_decimals))

    @staticmethod
    def ui_base_to_lots(perp_market: PerpMarket, ui_base: float) -> int:
        return int(Decimal(ui_base * 10 ** perp_market.base_decimals) / Decimal(perp_market.base_lot_size))

    @staticmethod
    def ui_quote_to_lots(perp_market: PerpMarket, ui_quote: float) -> int:
        return int(Decimal(ui_quote * 10 ** QUOTE_DECIMALS) / Decimal(perp_market.quote_lot_size))



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class TokenLiqBankruptcyArgs(typing.TypedDict):
    max_liab_transfer: types.i80f48.I80F48


layout = borsh.CStruct("max_liab_transfer" / types.i80f48.I80F48.layout)


class TokenLiqBankruptcyAccounts(typing.TypedDict):
    group: PublicKey
    liqor: PublicKey
    liqor_owner: PublicKey
    liqee: PublicKey
    liab_mint_info: PublicKey
    quote_vault: PublicKey
    insurance_vault: PublicKey


def token_liq_bankruptcy(
    args: TokenLiqBankruptcyArgs,
    accounts: TokenLiqBankruptcyAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["liqor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["liqor_owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["liqee"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["liab_mint_info"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"zn\xcb\x0f\x08u\xa4F"
    encoded_args = layout.build(
        {
            "max_liab_transfer": args["max_liab_transfer"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from mango_explorer_v4.types.token_position import TokenPosition
from mango_explorer_v4.accounts.bank import Bank
from decimal import Decimal


class TokenPositionHelper():
    @staticmethod
    def is_active(token_position: TokenPosition):
        return token_position.token_index != 65535

    @staticmethod
    def balance(token_position: TokenPosition, bank: Bank) -> Decimal:
        token_position = token_position.indexed_position.to_decimal()

        bank_index = bank.deposit_index if token_position > 0 else bank.borrow_index

        return (bank_index.to_decimal() * token_position) / Decimal(str(10 ** bank.mint_decimals))


from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class PerpOpenOrderJSON(typing.TypedDict):
    side_and_tree: int
    padding1: list[int]
    market: int
    padding2: list[int]
    client_id: int
    id: int
    reserved: list[int]


@dataclass
class PerpOpenOrder:
    layout: typing.ClassVar = borsh.CStruct(
        "side_and_tree" / borsh.U8,
        "padding1" / borsh.U8[1],
        "market" / borsh.U16,
        "padding2" / borsh.U8[4],
        "client_id" / borsh.U64,
        "id" / borsh.U128,
        "reserved" / borsh.U8[64],
    )
    side_and_tree: int
    padding1: list[int]
    market: int
    padding2: list[int]
    client_id: int
    id: int
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "PerpOpenOrder":
        return cls(
            side_and_tree=obj.side_and_tree,
            padding1=obj.padding1,
            market=obj.market,
            padding2=obj.padding2,
            client_id=obj.client_id,
            id=obj.id,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "side_and_tree": self.side_and_tree,
            "padding1": self.padding1,
            "market": self.market,
            "padding2": self.padding2,
            "client_id": self.client_id,
            "id": self.id,
            "reserved": self.reserved,
        }

    def to_json(self) -> PerpOpenOrderJSON:
        return {
            "side_and_tree": self.side_and_tree,
            "padding1": self.padding1,
            "market": self.market,
            "padding2": self.padding2,
            "client_id": self.client_id,
            "id": self.id,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: PerpOpenOrderJSON) -> "PerpOpenOrder":
        return cls(
            side_and_tree=obj["side_and_tree"],
            padding1=obj["padding1"],
            market=obj["market"],
            padding2=obj["padding2"],
            client_id=obj["client_id"],
            id=obj["id"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solana.publickey import PublicKey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class FillEventJSON(typing.TypedDict):
    event_type: int
    taker_side: int
    maker_out: int
    maker_slot: int
    padding: list[int]
    timestamp: int
    seq_num: int
    maker: str
    padding2: list[int]
    maker_timestamp: int
    taker: str
    padding3: list[int]
    taker_client_order_id: int
    padding4: list[int]
    price: int
    quantity: int
    maker_client_order_id: int
    maker_fee: float
    taker_fee: float
    reserved: list[int]


@dataclass
class FillEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "event_type" / borsh.U8,
        "taker_side" / borsh.U8,
        "maker_out" / borsh.U8,
        "maker_slot" / borsh.U8,
        "padding" / borsh.U8[4],
        "timestamp" / borsh.U64,
        "seq_num" / borsh.U64,
        "maker" / BorshPubkey,
        "padding2" / borsh.U8[32],
        "maker_timestamp" / borsh.U64,
        "taker" / BorshPubkey,
        "padding3" / borsh.U8[16],
        "taker_client_order_id" / borsh.U64,
        "padding4" / borsh.U8[16],
        "price" / borsh.I64,
        "quantity" / borsh.I64,
        "maker_client_order_id" / borsh.U64,
        "maker_fee" / borsh.F32,
        "taker_fee" / borsh.F32,
        "reserved" / borsh.U8[8],
    )
    event_type: int
    taker_side: int
    maker_out: int
    maker_slot: int
    padding: list[int]
    timestamp: int
    seq_num: int
    maker: PublicKey
    padding2: list[int]
    maker_timestamp: int
    taker: PublicKey
    padding3: list[int]
    taker_client_order_id: int
    padding4: list[int]
    price: int
    quantity: int
    maker_client_order_id: int
    maker_fee: float
    taker_fee: float
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "FillEvent":
        return cls(
            event_type=obj.event_type,
            taker_side=obj.taker_side,
            maker_out=obj.maker_out,
            maker_slot=obj.maker_slot,
            padding=obj.padding,
            timestamp=obj.timestamp,
            seq_num=obj.seq_num,
            maker=obj.maker,
            padding2=obj.padding2,
            maker_timestamp=obj.maker_timestamp,
            taker=obj.taker,
            padding3=obj.padding3,
            taker_client_order_id=obj.taker_client_order_id,
            padding4=obj.padding4,
            price=obj.price,
            quantity=obj.quantity,
            maker_client_order_id=obj.maker_client_order_id,
            maker_fee=obj.maker_fee,
            taker_fee=obj.taker_fee,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "event_type": self.event_type,
            "taker_side": self.taker_side,
            "maker_out": self.maker_out,
            "maker_slot": self.maker_slot,
            "padding": self.padding,
            "timestamp": self.timestamp,
            "seq_num": self.seq_num,
            "maker": self.maker,
            "padding2": self.padding2,
            "maker_timestamp": self.maker_timestamp,
            "taker": self.taker,
            "padding3": self.padding3,
            "taker_client_order_id": self.taker_client_order_id,
            "padding4": self.padding4,
            "price": self.price,
            "quantity": self.quantity,
            "maker_client_order_id": self.maker_client_order_id,
            "maker_fee": self.maker_fee,
            "taker_fee": self.taker_fee,
            "reserved": self.reserved,
        }

    def to_json(self) -> FillEventJSON:
        return {
            "event_type": self.event_type,
            "taker_side": self.taker_side,
            "maker_out": self.maker_out,
            "maker_slot": self.maker_slot,
            "padding": self.padding,
            "timestamp": self.timestamp,
            "seq_num": self.seq_num,
            "maker": str(self.maker),
            "padding2": self.padding2,
            "maker_timestamp": self.maker_timestamp,
            "taker": str(self.taker),
            "padding3": self.padding3,
            "taker_client_order_id": self.taker_client_order_id,
            "padding4": self.padding4,
            "price": self.price,
            "quantity": self.quantity,
            "maker_client_order_id": self.maker_client_order_id,
            "maker_fee": self.maker_fee,
            "taker_fee": self.taker_fee,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: FillEventJSON) -> "FillEvent":
        return cls(
            event_type=obj["event_type"],
            taker_side=obj["taker_side"],
            maker_out=obj["maker_out"],
            maker_slot=obj["maker_slot"],
            padding=obj["padding"],
            timestamp=obj["timestamp"],
            seq_num=obj["seq_num"],
            maker=PublicKey(obj["maker"]),
            padding2=obj["padding2"],
            maker_timestamp=obj["maker_timestamp"],
            taker=PublicKey(obj["taker"]),
            padding3=obj["padding3"],
            taker_client_order_id=obj["taker_client_order_id"],
            padding4=obj["padding4"],
            price=obj["price"],
            quantity=obj["quantity"],
            maker_client_order_id=obj["maker_client_order_id"],
            maker_fee=obj["maker_fee"],
            taker_fee=obj["taker_fee"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class PerpSettlePnlAccounts(typing.TypedDict):
    group: PublicKey
    settler: PublicKey
    settler_owner: PublicKey
    perp_market: PublicKey
    account_a: PublicKey
    account_b: PublicKey
    oracle: PublicKey
    settle_bank: PublicKey
    settle_oracle: PublicKey


def perp_settle_pnl(
    accounts: PerpSettlePnlAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["settler"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["settler_owner"], is_signer=True, is_writable=False
        ),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account_a"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["account_b"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["settle_bank"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["settle_oracle"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf5bU\xb3\xe6\xd7\x829"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solana.publickey import PublicKey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class MangoAccountFixedJSON(typing.TypedDict):
    group: str
    owner: str
    name: list[int]
    delegate: str
    account_num: int
    being_liquidated: int
    in_health_region: int
    bump: int
    padding: list[int]
    net_deposits: int
    perp_spot_transfers: int
    health_region_begin_init_health: int
    frozen_until: int
    buyback_fees_accrued_current: int
    buyback_fees_accrued_previous: int
    buyback_fees_expiry_timestamp: int
    reserved: list[int]


@dataclass
class MangoAccountFixed:
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "owner" / BorshPubkey,
        "name" / borsh.U8[32],
        "delegate" / BorshPubkey,
        "account_num" / borsh.U32,
        "being_liquidated" / borsh.U8,
        "in_health_region" / borsh.U8,
        "bump" / borsh.U8,
        "padding" / borsh.U8[1],
        "net_deposits" / borsh.I64,
        "perp_spot_transfers" / borsh.I64,
        "health_region_begin_init_health" / borsh.I64,
        "frozen_until" / borsh.U64,
        "buyback_fees_accrued_current" / borsh.U64,
        "buyback_fees_accrued_previous" / borsh.U64,
        "buyback_fees_expiry_timestamp" / borsh.U64,
        "reserved" / borsh.U8[208],
    )
    group: PublicKey
    owner: PublicKey
    name: list[int]
    delegate: PublicKey
    account_num: int
    being_liquidated: int
    in_health_region: int
    bump: int
    padding: list[int]
    net_deposits: int
    perp_spot_transfers: int
    health_region_begin_init_health: int
    frozen_until: int
    buyback_fees_accrued_current: int
    buyback_fees_accrued_previous: int
    buyback_fees_expiry_timestamp: int
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "MangoAccountFixed":
        return cls(
            group=obj.group,
            owner=obj.owner,
            name=obj.name,
            delegate=obj.delegate,
            account_num=obj.account_num,
            being_liquidated=obj.being_liquidated,
            in_health_region=obj.in_health_region,
            bump=obj.bump,
            padding=obj.padding,
            net_deposits=obj.net_deposits,
            perp_spot_transfers=obj.perp_spot_transfers,
            health_region_begin_init_health=obj.health_region_begin_init_health,
            frozen_until=obj.frozen_until,
            buyback_fees_accrued_current=obj.buyback_fees_accrued_current,
            buyback_fees_accrued_previous=obj.buyback_fees_accrued_previous,
            buyback_fees_expiry_timestamp=obj.buyback_fees_expiry_timestamp,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "group": self.group,
            "owner": self.owner,
            "name": self.name,
            "delegate": self.delegate,
            "account_num": self.account_num,
            "being_liquidated": self.being_liquidated,
            "in_health_region": self.in_health_region,
            "bump": self.bump,
            "padding": self.padding,
            "net_deposits": self.net_deposits,
            "perp_spot_transfers": self.perp_spot_transfers,
            "health_region_begin_init_health": self.health_region_begin_init_health,
            "frozen_until": self.frozen_until,
            "buyback_fees_accrued_current": self.buyback_fees_accrued_current,
            "buyback_fees_accrued_previous": self.buyback_fees_accrued_previous,
            "buyback_fees_expiry_timestamp": self.buyback_fees_expiry_timestamp,
            "reserved": self.reserved,
        }

    def to_json(self) -> MangoAccountFixedJSON:
        return {
            "group": str(self.group),
            "owner": str(self.owner),
            "name": self.name,
            "delegate": str(self.delegate),
            "account_num": self.account_num,
            "being_liquidated": self.being_liquidated,
            "in_health_region": self.in_health_region,
            "bump": self.bump,
            "padding": self.padding,
            "net_deposits": self.net_deposits,
            "perp_spot_transfers": self.perp_spot_transfers,
            "health_region_begin_init_health": self.health_region_begin_init_health,
            "frozen_until": self.frozen_until,
            "buyback_fees_accrued_current": self.buyback_fees_accrued_current,
            "buyback_fees_accrued_previous": self.buyback_fees_accrued_previous,
            "buyback_fees_expiry_timestamp": self.buyback_fees_expiry_timestamp,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: MangoAccountFixedJSON) -> "MangoAccountFixed":
        return cls(
            group=PublicKey(obj["group"]),
            owner=PublicKey(obj["owner"]),
            name=obj["name"],
            delegate=PublicKey(obj["delegate"]),
            account_num=obj["account_num"],
            being_liquidated=obj["being_liquidated"],
            in_health_region=obj["in_health_region"],
            bump=obj["bump"],
            padding=obj["padding"],
            net_deposits=obj["net_deposits"],
            perp_spot_transfers=obj["perp_spot_transfers"],
            health_region_begin_init_health=obj["health_region_begin_init_health"],
            frozen_until=obj["frozen_until"],
            buyback_fees_accrued_current=obj["buyback_fees_accrued_current"],
            buyback_fees_accrued_previous=obj["buyback_fees_accrued_previous"],
            buyback_fees_expiry_timestamp=obj["buyback_fees_expiry_timestamp"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class TokenEditArgs(typing.TypedDict):
    oracle_opt: typing.Optional[PublicKey]
    oracle_config_opt: typing.Optional[types.oracle_config_params.OracleConfigParams]
    group_insurance_fund_opt: typing.Optional[bool]
    interest_rate_params_opt: typing.Optional[
        types.interest_rate_params.InterestRateParams
    ]
    loan_fee_rate_opt: typing.Optional[float]
    loan_origination_fee_rate_opt: typing.Optional[float]
    maint_asset_weight_opt: typing.Optional[float]
    init_asset_weight_opt: typing.Optional[float]
    maint_liab_weight_opt: typing.Optional[float]
    init_liab_weight_opt: typing.Optional[float]
    liquidation_fee_opt: typing.Optional[float]
    stable_price_delay_interval_seconds_opt: typing.Optional[int]
    stable_price_delay_growth_limit_opt: typing.Optional[float]
    stable_price_growth_limit_opt: typing.Optional[float]
    min_vault_to_deposits_ratio_opt: typing.Optional[float]
    net_borrow_limit_per_window_quote_opt: typing.Optional[int]
    net_borrow_limit_window_size_ts_opt: typing.Optional[int]
    borrow_weight_scale_start_quote_opt: typing.Optional[float]
    deposit_weight_scale_start_quote_opt: typing.Optional[float]
    reset_stable_price: bool
    reset_net_borrow_limit: bool
    reduce_only_opt: typing.Optional[bool]
    name_opt: typing.Optional[str]


layout = borsh.CStruct(
    "oracle_opt" / borsh.Option(BorshPubkey),
    "oracle_config_opt"
    / borsh.Option(types.oracle_config_params.OracleConfigParams.layout),
    "group_insurance_fund_opt" / borsh.Option(borsh.Bool),
    "interest_rate_params_opt"
    / borsh.Option(types.interest_rate_params.InterestRateParams.layout),
    "loan_fee_rate_opt" / borsh.Option(borsh.F32),
    "loan_origination_fee_rate_opt" / borsh.Option(borsh.F32),
    "maint_asset_weight_opt" / borsh.Option(borsh.F32),
    "init_asset_weight_opt" / borsh.Option(borsh.F32),
    "maint_liab_weight_opt" / borsh.Option(borsh.F32),
    "init_liab_weight_opt" / borsh.Option(borsh.F32),
    "liquidation_fee_opt" / borsh.Option(borsh.F32),
    "stable_price_delay_interval_seconds_opt" / borsh.Option(borsh.U32),
    "stable_price_delay_growth_limit_opt" / borsh.Option(borsh.F32),
    "stable_price_growth_limit_opt" / borsh.Option(borsh.F32),
    "min_vault_to_deposits_ratio_opt" / borsh.Option(borsh.F64),
    "net_borrow_limit_per_window_quote_opt" / borsh.Option(borsh.I64),
    "net_borrow_limit_window_size_ts_opt" / borsh.Option(borsh.U64),
    "borrow_weight_scale_start_quote_opt" / borsh.Option(borsh.F64),
    "deposit_weight_scale_start_quote_opt" / borsh.Option(borsh.F64),
    "reset_stable_price" / borsh.Bool,
    "reset_net_borrow_limit" / borsh.Bool,
    "reduce_only_opt" / borsh.Option(borsh.Bool),
    "name_opt" / borsh.Option(borsh.String),
)


class TokenEditAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    mint_info: PublicKey
    oracle: PublicKey


def token_edit(
    args: TokenEditArgs,
    accounts: TokenEditAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["mint_info"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x91\xcc\x0b\xd1\xae\x86O>"
    encoded_args = layout.build(
        {
            "oracle_opt": args["oracle_opt"],
            "oracle_config_opt": (
                None
                if args["oracle_config_opt"] is None
                else args["oracle_config_opt"].to_encodable()
            ),
            "group_insurance_fund_opt": args["group_insurance_fund_opt"],
            "interest_rate_params_opt": (
                None
                if args["interest_rate_params_opt"] is None
                else args["interest_rate_params_opt"].to_encodable()
            ),
            "loan_fee_rate_opt": args["loan_fee_rate_opt"],
            "loan_origination_fee_rate_opt": args["loan_origination_fee_rate_opt"],
            "maint_asset_weight_opt": args["maint_asset_weight_opt"],
            "init_asset_weight_opt": args["init_asset_weight_opt"],
            "maint_liab_weight_opt": args["maint_liab_weight_opt"],
            "init_liab_weight_opt": args["init_liab_weight_opt"],
            "liquidation_fee_opt": args["liquidation_fee_opt"],
            "stable_price_delay_interval_seconds_opt": args[
                "stable_price_delay_interval_seconds_opt"
            ],
            "stable_price_delay_growth_limit_opt": args[
                "stable_price_delay_growth_limit_opt"
            ],
            "stable_price_growth_limit_opt": args["stable_price_growth_limit_opt"],
            "min_vault_to_deposits_ratio_opt": args["min_vault_to_deposits_ratio_opt"],
            "net_borrow_limit_per_window_quote_opt": args[
                "net_borrow_limit_per_window_quote_opt"
            ],
            "net_borrow_limit_window_size_ts_opt": args[
                "net_borrow_limit_window_size_ts_opt"
            ],
            "borrow_weight_scale_start_quote_opt": args[
                "borrow_weight_scale_start_quote_opt"
            ],
            "deposit_weight_scale_start_quote_opt": args[
                "deposit_weight_scale_start_quote_opt"
            ],
            "reset_stable_price": args["reset_stable_price"],
            "reset_net_borrow_limit": args["reset_net_borrow_limit"],
            "reduce_only_opt": args["reduce_only_opt"],
            "name_opt": args["name_opt"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import typing
from . import interest_rate_params
from .interest_rate_params import InterestRateParams, InterestRateParamsJSON
from . import equity
from .equity import Equity, EquityJSON
from . import token_equity
from .token_equity import TokenEquity, TokenEquityJSON
from . import perp_equity
from .perp_equity import PerpEquity, PerpEquityJSON
from . import prices
from .prices import Prices, PricesJSON
from . import token_info
from .token_info import TokenInfo, TokenInfoJSON
from . import serum3_info
from .serum3_info import Serum3Info, Serum3InfoJSON
from . import perp_info
from .perp_info import PerpInfo, PerpInfoJSON
from . import health_cache
from .health_cache import HealthCache, HealthCacheJSON
from . import flash_loan_token_detail
from .flash_loan_token_detail import FlashLoanTokenDetail, FlashLoanTokenDetailJSON
from . import token_position
from .token_position import TokenPosition, TokenPositionJSON
from . import serum3_orders
from .serum3_orders import Serum3Orders, Serum3OrdersJSON
from . import perp_position
from .perp_position import PerpPosition, PerpPositionJSON
from . import perp_open_order
from .perp_open_order import PerpOpenOrder, PerpOpenOrderJSON
from . import mango_account_fixed
from .mango_account_fixed import MangoAccountFixed, MangoAccountFixedJSON
from . import oracle_config
from .oracle_config import OracleConfig, OracleConfigJSON
from . import oracle_config_params
from .oracle_config_params import OracleConfigParams, OracleConfigParamsJSON
from . import inner_node
from .inner_node import InnerNode, InnerNodeJSON
from . import leaf_node
from .leaf_node import LeafNode, LeafNodeJSON
from . import any_node
from .any_node import AnyNode, AnyNodeJSON
from . import order_tree_root
from .order_tree_root import OrderTreeRoot, OrderTreeRootJSON
from . import order_tree_nodes
from .order_tree_nodes import OrderTreeNodes, OrderTreeNodesJSON
from . import event_queue_header
from .event_queue_header import EventQueueHeader, EventQueueHeaderJSON
from . import any_event
from .any_event import AnyEvent, AnyEventJSON
from . import fill_event
from .fill_event import FillEvent, FillEventJSON
from . import out_event
from .out_event import OutEvent, OutEventJSON
from . import stable_price_model
from .stable_price_model import StablePriceModel, StablePriceModelJSON
from . import token_index
from .token_index import TokenIndex, TokenIndexJSON
from . import serum3_market_index
from .serum3_market_index import Serum3MarketIndex, Serum3MarketIndexJSON
from . import perp_market_index
from .perp_market_index import PerpMarketIndex, PerpMarketIndexJSON
from . import i80f48
from .i80f48 import I80F48, I80F48JSON
from . import flash_loan_type
from .flash_loan_type import FlashLoanTypeKind, FlashLoanTypeJSON
from . import serum3_self_trade_behavior
from .serum3_self_trade_behavior import (
    Serum3SelfTradeBehaviorKind,
    Serum3SelfTradeBehaviorJSON,
)
from . import serum3_order_type
from .serum3_order_type import Serum3OrderTypeKind, Serum3OrderTypeJSON
from . import serum3_side
from .serum3_side import Serum3SideKind, Serum3SideJSON
from . import health_type
from .health_type import HealthTypeKind, HealthTypeJSON
from . import loan_origination_fee_instruction
from .loan_origination_fee_instruction import (
    LoanOriginationFeeInstructionKind,
    LoanOriginationFeeInstructionJSON,
)
from . import ix_gate
from .ix_gate import IxGateKind, IxGateJSON
from . import oracle_type
from .oracle_type import OracleTypeKind, OracleTypeJSON
from . import order_state
from .order_state import OrderStateKind, OrderStateJSON
from . import book_side_order_tree
from .book_side_order_tree import BookSideOrderTreeKind, BookSideOrderTreeJSON
from . import node_tag
from .node_tag import NodeTagKind, NodeTagJSON
from . import place_order_type
from .place_order_type import PlaceOrderTypeKind, PlaceOrderTypeJSON
from . import post_order_type
from .post_order_type import PostOrderTypeKind, PostOrderTypeJSON
from . import side
from .side import SideKind, SideJSON
from . import side_and_order_tree
from .side_and_order_tree import SideAndOrderTreeKind, SideAndOrderTreeJSON
from . import order_params
from .order_params import OrderParamsKind, OrderParamsJSON
from . import order_tree_type
from .order_tree_type import OrderTreeTypeKind, OrderTreeTypeJSON
from . import event_type
from .event_type import EventTypeKind, EventTypeJSON



from mango_explorer_v4.accounts.mango_account import MangoAccount
from mango_explorer_v4.types.health_cache import HealthCache
from mango_explorer_v4.helpers.mango_account import MangoAccountHelper


class HealthCacheHelper():
    @staticmethod
    def from_mango_account(mango_account: MangoAccount):
        token_infos = [
            token_position.token_index
            for token_position in  MangoAccountHelper.active_token_positions(mango_account)
        ]


from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class StablePriceModelJSON(typing.TypedDict):
    stable_price: float
    last_update_timestamp: int
    delay_prices: list[float]
    delay_accumulator_price: float
    delay_accumulator_time: int
    delay_interval_seconds: int
    delay_growth_limit: float
    stable_growth_limit: float
    last_delay_interval_index: int
    padding: list[int]
    reserved: list[int]


@dataclass
class StablePriceModel:
    layout: typing.ClassVar = borsh.CStruct(
        "stable_price" / borsh.F64,
        "last_update_timestamp" / borsh.U64,
        "delay_prices" / borsh.F64[24],
        "delay_accumulator_price" / borsh.F64,
        "delay_accumulator_time" / borsh.U32,
        "delay_interval_seconds" / borsh.U32,
        "delay_growth_limit" / borsh.F32,
        "stable_growth_limit" / borsh.F32,
        "last_delay_interval_index" / borsh.U8,
        "padding" / borsh.U8[7],
        "reserved" / borsh.U8[48],
    )
    stable_price: float
    last_update_timestamp: int
    delay_prices: list[float]
    delay_accumulator_price: float
    delay_accumulator_time: int
    delay_interval_seconds: int
    delay_growth_limit: float
    stable_growth_limit: float
    last_delay_interval_index: int
    padding: list[int]
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "StablePriceModel":
        return cls(
            stable_price=obj.stable_price,
            last_update_timestamp=obj.last_update_timestamp,
            delay_prices=obj.delay_prices,
            delay_accumulator_price=obj.delay_accumulator_price,
            delay_accumulator_time=obj.delay_accumulator_time,
            delay_interval_seconds=obj.delay_interval_seconds,
            delay_growth_limit=obj.delay_growth_limit,
            stable_growth_limit=obj.stable_growth_limit,
            last_delay_interval_index=obj.last_delay_interval_index,
            padding=obj.padding,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "stable_price": self.stable_price,
            "last_update_timestamp": self.last_update_timestamp,
            "delay_prices": self.delay_prices,
            "delay_accumulator_price": self.delay_accumulator_price,
            "delay_accumulator_time": self.delay_accumulator_time,
            "delay_interval_seconds": self.delay_interval_seconds,
            "delay_growth_limit": self.delay_growth_limit,
            "stable_growth_limit": self.stable_growth_limit,
            "last_delay_interval_index": self.last_delay_interval_index,
            "padding": self.padding,
            "reserved": self.reserved,
        }

    def to_json(self) -> StablePriceModelJSON:
        return {
            "stable_price": self.stable_price,
            "last_update_timestamp": self.last_update_timestamp,
            "delay_prices": self.delay_prices,
            "delay_accumulator_price": self.delay_accumulator_price,
            "delay_accumulator_time": self.delay_accumulator_time,
            "delay_interval_seconds": self.delay_interval_seconds,
            "delay_growth_limit": self.delay_growth_limit,
            "stable_growth_limit": self.stable_growth_limit,
            "last_delay_interval_index": self.last_delay_interval_index,
            "padding": self.padding,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: StablePriceModelJSON) -> "StablePriceModel":
        return cls(
            stable_price=obj["stable_price"],
            last_update_timestamp=obj["last_update_timestamp"],
            delay_prices=obj["delay_prices"],
            delay_accumulator_price=obj["delay_accumulator_price"],
            delay_accumulator_time=obj["delay_accumulator_time"],
            delay_interval_seconds=obj["delay_interval_seconds"],
            delay_growth_limit=obj["delay_growth_limit"],
            stable_growth_limit=obj["stable_growth_limit"],
            last_delay_interval_index=obj["last_delay_interval_index"],
            padding=obj["padding"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class OrderTreeRootJSON(typing.TypedDict):
    maybe_node: int
    leaf_count: int


@dataclass
class OrderTreeRoot:
    layout: typing.ClassVar = borsh.CStruct(
        "maybe_node" / borsh.U32, "leaf_count" / borsh.U32
    )
    maybe_node: int
    leaf_count: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "OrderTreeRoot":
        return cls(maybe_node=obj.maybe_node, leaf_count=obj.leaf_count)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"maybe_node": self.maybe_node, "leaf_count": self.leaf_count}

    def to_json(self) -> OrderTreeRootJSON:
        return {"maybe_node": self.maybe_node, "leaf_count": self.leaf_count}

    @classmethod
    def from_json(cls, obj: OrderTreeRootJSON) -> "OrderTreeRoot":
        return cls(maybe_node=obj["maybe_node"], leaf_count=obj["leaf_count"])



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class Serum3RegisterMarketArgs(typing.TypedDict):
    market_index: int
    name: str


layout = borsh.CStruct("market_index" / borsh.U16, "name" / borsh.String)


class Serum3RegisterMarketAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    serum_market: PublicKey
    index_reservation: PublicKey
    quote_bank: PublicKey
    base_bank: PublicKey
    payer: PublicKey


def serum3_register_market(
    args: Serum3RegisterMarketArgs,
    accounts: Serum3RegisterMarketAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_market_external"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["serum_market"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["index_reservation"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["quote_bank"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["base_bank"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"(\x0emx\xde\x9c\xd1\x00"
    encoded_args = layout.build(
        {
            "market_index": args["market_index"],
            "name": args["name"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FlashLoanTokenDetailJSON(typing.TypedDict):
    token_index: int
    change_amount: int
    loan: int
    loan_origination_fee: int
    deposit_index: int
    borrow_index: int
    price: int


@dataclass
class FlashLoanTokenDetail:
    layout: typing.ClassVar = borsh.CStruct(
        "token_index" / borsh.U16,
        "change_amount" / borsh.I128,
        "loan" / borsh.I128,
        "loan_origination_fee" / borsh.I128,
        "deposit_index" / borsh.I128,
        "borrow_index" / borsh.I128,
        "price" / borsh.I128,
    )
    token_index: int
    change_amount: int
    loan: int
    loan_origination_fee: int
    deposit_index: int
    borrow_index: int
    price: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "FlashLoanTokenDetail":
        return cls(
            token_index=obj.token_index,
            change_amount=obj.change_amount,
            loan=obj.loan,
            loan_origination_fee=obj.loan_origination_fee,
            deposit_index=obj.deposit_index,
            borrow_index=obj.borrow_index,
            price=obj.price,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "token_index": self.token_index,
            "change_amount": self.change_amount,
            "loan": self.loan,
            "loan_origination_fee": self.loan_origination_fee,
            "deposit_index": self.deposit_index,
            "borrow_index": self.borrow_index,
            "price": self.price,
        }

    def to_json(self) -> FlashLoanTokenDetailJSON:
        return {
            "token_index": self.token_index,
            "change_amount": self.change_amount,
            "loan": self.loan,
            "loan_origination_fee": self.loan_origination_fee,
            "deposit_index": self.deposit_index,
            "borrow_index": self.borrow_index,
            "price": self.price,
        }

    @classmethod
    def from_json(cls, obj: FlashLoanTokenDetailJSON) -> "FlashLoanTokenDetail":
        return cls(
            token_index=obj["token_index"],
            change_amount=obj["change_amount"],
            loan=obj["loan"],
            loan_origination_fee=obj["loan_origination_fee"],
            deposit_index=obj["deposit_index"],
            borrow_index=obj["borrow_index"],
            price=obj["price"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class PerpCancelAllOrdersBySideArgs(typing.TypedDict):
    side_option: typing.Optional[types.side.SideKind]
    limit: int


layout = borsh.CStruct(
    "side_option" / borsh.Option(types.side.layout), "limit" / borsh.U8
)


class PerpCancelAllOrdersBySideAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey


def perp_cancel_all_orders_by_side(
    args: PerpCancelAllOrdersBySideArgs,
    accounts: PerpCancelAllOrdersBySideAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"3\xf8\xcc}e\xb6k\x92"
    encoded_args = layout.build(
        {
            "side_option": (
                None
                if args["side_option"] is None
                else args["side_option"].to_encodable()
            ),
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class PerpSettleFeesArgs(typing.TypedDict):
    max_settle_amount: int


layout = borsh.CStruct("max_settle_amount" / borsh.U64)


class PerpSettleFeesAccounts(typing.TypedDict):
    group: PublicKey
    perp_market: PublicKey
    account: PublicKey
    oracle: PublicKey
    settle_bank: PublicKey
    settle_oracle: PublicKey


def perp_settle_fees(
    args: PerpSettleFeesArgs,
    accounts: PerpSettleFeesAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["settle_bank"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["settle_oracle"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdf\xed\xe3H\x98\xb9\xeas"
    encoded_args = layout.build(
        {
            "max_settle_amount": args["max_settle_amount"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class BidFixedJSON(typing.TypedDict):
    kind: typing.Literal["BidFixed"]


class AskFixedJSON(typing.TypedDict):
    kind: typing.Literal["AskFixed"]


class BidOraclePeggedJSON(typing.TypedDict):
    kind: typing.Literal["BidOraclePegged"]


class AskOraclePeggedJSON(typing.TypedDict):
    kind: typing.Literal["AskOraclePegged"]


@dataclass
class BidFixed:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "BidFixed"

    @classmethod
    def to_json(cls) -> BidFixedJSON:
        return BidFixedJSON(
            kind="BidFixed",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BidFixed": {},
        }


@dataclass
class AskFixed:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "AskFixed"

    @classmethod
    def to_json(cls) -> AskFixedJSON:
        return AskFixedJSON(
            kind="AskFixed",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AskFixed": {},
        }


@dataclass
class BidOraclePegged:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "BidOraclePegged"

    @classmethod
    def to_json(cls) -> BidOraclePeggedJSON:
        return BidOraclePeggedJSON(
            kind="BidOraclePegged",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BidOraclePegged": {},
        }


@dataclass
class AskOraclePegged:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "AskOraclePegged"

    @classmethod
    def to_json(cls) -> AskOraclePeggedJSON:
        return AskOraclePeggedJSON(
            kind="AskOraclePegged",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AskOraclePegged": {},
        }


SideAndOrderTreeKind = typing.Union[
    BidFixed, AskFixed, BidOraclePegged, AskOraclePegged
]
SideAndOrderTreeJSON = typing.Union[
    BidFixedJSON, AskFixedJSON, BidOraclePeggedJSON, AskOraclePeggedJSON
]


def from_decoded(obj: dict) -> SideAndOrderTreeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "BidFixed" in obj:
        return BidFixed()
    if "AskFixed" in obj:
        return AskFixed()
    if "BidOraclePegged" in obj:
        return BidOraclePegged()
    if "AskOraclePegged" in obj:
        return AskOraclePegged()
    raise ValueError("Invalid enum object")


def from_json(obj: SideAndOrderTreeJSON) -> SideAndOrderTreeKind:
    if obj["kind"] == "BidFixed":
        return BidFixed()
    if obj["kind"] == "AskFixed":
        return AskFixed()
    if obj["kind"] == "BidOraclePegged":
        return BidOraclePegged()
    if obj["kind"] == "AskOraclePegged":
        return AskOraclePegged()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "BidFixed" / borsh.CStruct(),
    "AskFixed" / borsh.CStruct(),
    "BidOraclePegged" / borsh.CStruct(),
    "AskOraclePegged" / borsh.CStruct(),
)



import decimal

from mango_explorer_v4.types.prices import Prices
from mango_explorer_v4.types.health_type import HealthTypeKind, Maint, LiquidationEnd


class PricesHelper:
    @staticmethod
    def liab(prices: Prices, health_type: HealthTypeKind) -> decimal.Decimal:
        if health_type in [Maint, LiquidationEnd] or health_type is None:
            return prices.oracle.to_decimal()
        else:
            return max(prices.oracle.to_decimal(), prices.stable.to_decimal())

    @staticmethod
    def asset(prices: Prices, health_type: HealthTypeKind) -> decimal.Decimal:
        if health_type in [Maint, LiquidationEnd] or health_type is None:
            return prices.oracle.to_decimal()
        else:
            return min(prices.oracle.to_decimal(), prices.stable.to_decimal())



from decimal import Decimal
from mango_explorer_v4.types.token_info import TokenInfo
from mango_explorer_v4.types.health_type import HealthTypeKind, Init, Maint, LiquidationEnd
from mango_explorer_v4.helpers.prices import PricesHelper


class TokenInfoHelper():
    @staticmethod
    def liab_weight(token_info: TokenInfo, health_type: HealthTypeKind) -> Decimal:
        return {
            Init: token_info.init_scaled_liab_weight.to_decimal(),
            Maint: token_info.maint_liab_weight.to_decimal(),
            LiquidationEnd: token_info.init_scaled_liab_weight.to_decimal()
        }[type(health_type)]

    @staticmethod
    def asset_weight(token_info: TokenInfo, health_type: HealthTypeKind) -> Decimal:
        return {
            Init: token_info.init_scaled_asset_weight.to_decimal(),
            Maint: token_info.maint_asset_weight.to_decimal(),
            LiquidationEnd: token_info.init_asset_weight.to_decimal()
        }[type(health_type)]

    @staticmethod
    def health_contribution(token_info: TokenInfo, health_type: HealthTypeKind):
        print(type(health_type), type(health_type) == Init, type(health_type))

        if health_type is None:
            return token_info.balance_native.to_decimal() * token_info.prices.oracle.to_decimal()

        if token_info.balance_native.to_decimal() < 0:
            weight, price = TokenInfoHelper.liab_weight(token_info, health_type), PricesHelper.liab(token_info.prices, health_type)
        else:
            weight, price = TokenInfoHelper.asset_weight(token_info, health_type), PricesHelper.asset(token_info.prices, health_type)

        return token_info.balance_native.to_decimal() * weight * price



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class ValidJSON(typing.TypedDict):
    kind: typing.Literal["Valid"]


class InvalidJSON(typing.TypedDict):
    kind: typing.Literal["Invalid"]


class SkippedJSON(typing.TypedDict):
    kind: typing.Literal["Skipped"]


@dataclass
class Valid:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Valid"

    @classmethod
    def to_json(cls) -> ValidJSON:
        return ValidJSON(
            kind="Valid",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Valid": {},
        }


@dataclass
class Invalid:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Invalid"

    @classmethod
    def to_json(cls) -> InvalidJSON:
        return InvalidJSON(
            kind="Invalid",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Invalid": {},
        }


@dataclass
class Skipped:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Skipped"

    @classmethod
    def to_json(cls) -> SkippedJSON:
        return SkippedJSON(
            kind="Skipped",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Skipped": {},
        }


OrderStateKind = typing.Union[Valid, Invalid, Skipped]
OrderStateJSON = typing.Union[ValidJSON, InvalidJSON, SkippedJSON]


def from_decoded(obj: dict) -> OrderStateKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Valid" in obj:
        return Valid()
    if "Invalid" in obj:
        return Invalid()
    if "Skipped" in obj:
        return Skipped()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderStateJSON) -> OrderStateKind:
    if obj["kind"] == "Valid":
        return Valid()
    if obj["kind"] == "Invalid":
        return Invalid()
    if obj["kind"] == "Skipped":
        return Skipped()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Valid" / borsh.CStruct(), "Invalid" / borsh.CStruct(), "Skipped" / borsh.CStruct()
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class LiqTokenWithTokenArgs(typing.TypedDict):
    asset_token_index: int
    liab_token_index: int
    max_liab_transfer: types.i80f48.I80F48


layout = borsh.CStruct(
    "asset_token_index" / borsh.U16,
    "liab_token_index" / borsh.U16,
    "max_liab_transfer" / types.i80f48.I80F48.layout,
)


class LiqTokenWithTokenAccounts(typing.TypedDict):
    group: PublicKey
    liqor: PublicKey
    liqor_owner: PublicKey
    liqee: PublicKey


def liq_token_with_token(
    args: LiqTokenWithTokenArgs,
    accounts: LiqTokenWithTokenAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["liqor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["liqor_owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["liqee"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"C\x7f\x98\x98\xd3\xd0\xfb\xe2"
    encoded_args = layout.build(
        {
            "asset_token_index": args["asset_token_index"],
            "liab_token_index": args["liab_token_index"],
            "max_liab_transfer": args["max_liab_transfer"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class InnerNodeJSON(typing.TypedDict):
    kind: typing.Literal["InnerNode"]


class LeafNodeJSON(typing.TypedDict):
    kind: typing.Literal["LeafNode"]


class FreeNodeJSON(typing.TypedDict):
    kind: typing.Literal["FreeNode"]


class LastFreeNodeJSON(typing.TypedDict):
    kind: typing.Literal["LastFreeNode"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class InnerNode:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "InnerNode"

    @classmethod
    def to_json(cls) -> InnerNodeJSON:
        return InnerNodeJSON(
            kind="InnerNode",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InnerNode": {},
        }


@dataclass
class LeafNode:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "LeafNode"

    @classmethod
    def to_json(cls) -> LeafNodeJSON:
        return LeafNodeJSON(
            kind="LeafNode",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LeafNode": {},
        }


@dataclass
class FreeNode:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "FreeNode"

    @classmethod
    def to_json(cls) -> FreeNodeJSON:
        return FreeNodeJSON(
            kind="FreeNode",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FreeNode": {},
        }


@dataclass
class LastFreeNode:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "LastFreeNode"

    @classmethod
    def to_json(cls) -> LastFreeNodeJSON:
        return LastFreeNodeJSON(
            kind="LastFreeNode",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LastFreeNode": {},
        }


NodeTagKind = typing.Union[Uninitialized, InnerNode, LeafNode, FreeNode, LastFreeNode]
NodeTagJSON = typing.Union[
    UninitializedJSON, InnerNodeJSON, LeafNodeJSON, FreeNodeJSON, LastFreeNodeJSON
]


def from_decoded(obj: dict) -> NodeTagKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "InnerNode" in obj:
        return InnerNode()
    if "LeafNode" in obj:
        return LeafNode()
    if "FreeNode" in obj:
        return FreeNode()
    if "LastFreeNode" in obj:
        return LastFreeNode()
    raise ValueError("Invalid enum object")


def from_json(obj: NodeTagJSON) -> NodeTagKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "InnerNode":
        return InnerNode()
    if obj["kind"] == "LeafNode":
        return LeafNode()
    if obj["kind"] == "FreeNode":
        return FreeNode()
    if obj["kind"] == "LastFreeNode":
        return LastFreeNode()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "InnerNode" / borsh.CStruct(),
    "LeafNode" / borsh.CStruct(),
    "FreeNode" / borsh.CStruct(),
    "LastFreeNode" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class TokenDepositArgs(typing.TypedDict):
    amount: int
    reduce_only: bool


layout = borsh.CStruct("amount" / borsh.U64, "reduce_only" / borsh.Bool)


class TokenDepositAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    bank: PublicKey
    vault: PublicKey
    oracle: PublicKey
    token_account: PublicKey
    token_authority: PublicKey


def token_deposit(
    args: TokenDepositArgs,
    accounts: TokenDepositAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["token_authority"], is_signer=True, is_writable=False
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"u\xff\x9aG\xf5:_Y"
    encoded_args = layout.build(
        {
            "amount": args["amount"],
            "reduce_only": args["reduce_only"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class AnyNodeJSON(typing.TypedDict):
    tag: int
    data: list[int]


@dataclass
class AnyNode:
    layout: typing.ClassVar = borsh.CStruct("tag" / borsh.U8, "data" / borsh.U8[119])
    tag: int
    data: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "AnyNode":
        return cls(tag=obj.tag, data=obj.data)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"tag": self.tag, "data": self.data}

    def to_json(self) -> AnyNodeJSON:
        return {"tag": self.tag, "data": self.data}

    @classmethod
    def from_json(cls, obj: AnyNodeJSON) -> "AnyNode":
        return cls(tag=obj["tag"], data=obj["data"])



import asyncio
import json
import logging
import pathlib
import re
import sys
import time
from dataclasses import dataclass
from decimal import Decimal
from typing import Literal

import aiostream.stream
import anchorpy.error
import base58
from anchorpy import Provider, Wallet
from pyserum.async_open_orders_account import AsyncOpenOrdersAccount
from pyserum.market import AsyncMarket, OrderBook
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Processed
from solana.rpc.websocket_api import connect
from solana.transaction import AccountMeta, Transaction
from solana.rpc.types import MemcmpOpts
from solders.rpc.responses import AccountNotification
from solders.account import Account

from mango_explorer_v4.accounts.bank import Bank
from mango_explorer_v4.accounts.book_side import BookSide
from mango_explorer_v4.accounts.event_queue import EventQueue
from mango_explorer_v4.accounts.mango_account import MangoAccount
from mango_explorer_v4.accounts.mint_info import MintInfo
from mango_explorer_v4.accounts.perp_market import PerpMarket
from mango_explorer_v4.accounts.serum3_market import Serum3Market
from mango_explorer_v4.helpers.serum3_orders import Serum3OrdersHelper
from mango_explorer_v4.helpers.token_position import TokenPositionHelper
from mango_explorer_v4.helpers.perp_market import PerpMarketHelper
from mango_explorer_v4.helpers.perp_position import PerpPositionHelper
from mango_explorer_v4.helpers.bank import BankHelper
from mango_explorer_v4.helpers.prices import PricesHelper
from mango_explorer_v4.helpers.mango_account import MangoAccountHelper
from mango_explorer_v4.helpers.token_info import TokenInfoHelper
from mango_explorer_v4.helpers.serum3_info import Serum3InfoHelper
from mango_explorer_v4.helpers.perp_info import PerpInfoHelper
from mango_explorer_v4.instructions.perp_cancel_all_orders import PerpCancelAllOrdersArgs, PerpCancelAllOrdersAccounts, perp_cancel_all_orders
from mango_explorer_v4.instructions.perp_place_order import PerpPlaceOrderArgs, PerpPlaceOrderAccounts, perp_place_order
from mango_explorer_v4.instructions.perp_place_order_pegged import PerpPlaceOrderPeggedArgs, PerpPlaceOrderPeggedAccounts, perp_place_order_pegged
from mango_explorer_v4.instructions.serum3_cancel_all_orders import Serum3CancelAllOrdersAccounts, Serum3CancelAllOrdersArgs, serum3_cancel_all_orders
from mango_explorer_v4.instructions.serum3_create_open_orders import Serum3CreateOpenOrdersAccounts, serum3_create_open_orders
from mango_explorer_v4.instructions.serum3_place_order import Serum3PlaceOrderArgs, Serum3PlaceOrderAccounts, serum3_place_order
from mango_explorer_v4.program_id import PROGRAM_ID as MANGO_PROGRAM_ID
from mango_explorer_v4.types import place_order_type
from mango_explorer_v4.types import serum3_side, serum3_self_trade_behavior, serum3_order_type
from mango_explorer_v4.types.fill_event import FillEvent
from mango_explorer_v4.types.place_order_type import Limit
from mango_explorer_v4.types.side import Bid, Ask
from mango_explorer_v4.types.token_info import TokenInfo
from mango_explorer_v4.types.perp_info import PerpInfo
from mango_explorer_v4.types.prices import Prices
from mango_explorer_v4.types.i80f48 import I80F48
from mango_explorer_v4.types.health_type import Init, Maint, LiquidationEnd
from mango_explorer_v4.types.serum3_info import Serum3Info
from mango_explorer_v4.types.health_cache import HealthCache
from mango_explorer_v4.types.health_type import HealthTypeKind
from .constants import RUST_I64_MAX, QUOTE_DECIMALS, SERUM_PROGRAM_ID
from .constructs.book_side_items import BookSideItems
from .constructs.serum3_reserved import Serum3Reserved
from .oracles import pyth

logging.basicConfig(
    level=logging.INFO
)


@dataclass
class MangoClient():
    connection: AsyncClient
    group_config: dict
    serum_market_configs: [dict]
    perp_market_configs: [dict]
    serum_markets: [Serum3Market]
    serum_markets_external: [AsyncMarket]
    perp_markets: [PerpMarket]
    banks: [Bank]
    mint_infos: [MintInfo]

    @staticmethod
    async def connect(rpc_url: str = 'https://mango.rpcpool.com/0f9acc0d45173b51bf7d7e09c1e5'):
        connection = AsyncClient(rpc_url, Processed)

        ids = json.loads(open(pathlib.Path(__file__).parent / 'ids.json').read())
        # TODO: ^ Make it fetch from https://api.mngo.cloud/data/v4/group-metadata instead

        group_config = [group_config for group_config in ids['groups'] if group_config['publicKey'] == '78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX'][0]
        # TODO: ^ Un-hardcode the group config public key (not necessary for now as there's only mainnet)

        perp_market_configs = [
            perp_market_config for perp_market_config in group_config['perpMarkets'] if perp_market_config['active']
        ]

        serum_market_configs = [
            serum_market_config for serum_market_config in group_config['serum3Markets'] if serum_market_config['active']
        ]

        bank_configs = [
            {
                'tokenIndex': token_config['tokenIndex'],
                'publicKey': PublicKey(token_config['banks'][0]['publicKey'])
            }
            for token_config in group_config['tokens']
            if token_config['active']
        ]

        mint_info_configs = [
            {
                'tokenIndex': token_config['tokenIndex'],
                'publicKey': PublicKey(token_config['mintInfo'])
            }
            for token_config in group_config['tokens']
            if token_config['active']
        ]

        mint_info_configs = list(sorted(mint_info_configs, key=lambda mint_info_config: mint_info_config['tokenIndex']))

        serum_markets = await Serum3Market.fetch_multiple(
            connection,
            [
                PublicKey(serum3_market_config['publicKey']) for serum3_market_config in serum_market_configs
            ]
        )

        serum_markets_external = await asyncio.gather(*[
            AsyncMarket.load(connection, serum_market.serum_market_external, SERUM_PROGRAM_ID)
            for serum_market in serum_markets
        ])

        banks, mint_infos, perp_markets = await asyncio.gather(*[
            Bank.fetch_multiple(
                connection,
                [bank_config['publicKey'] for bank_config in bank_configs]
            ),
            MintInfo.fetch_multiple(
                connection,
                [mint_info_config['publicKey'] for mint_info_config in mint_info_configs]
            ),
            PerpMarket.fetch_multiple(
                connection,
                [perp_market_config['publicKey'] for perp_market_config in perp_market_configs]
            )
        ])

        return MangoClient(
            connection=connection,
            group_config=group_config,
            serum_market_configs=serum_market_configs,
            perp_market_configs=perp_market_configs,
            serum_markets=serum_markets,
            serum_markets_external=serum_markets_external,
            perp_markets=perp_markets,
            banks=banks,
            mint_infos=mint_infos
        )

    def symbols(self):
        # This might not be the best format for keeping symbols organized,
        # as it presumes that perpetual and spot market names would never
        # conflict, but it works for now

        # TODO: Add minimum order size, tick size and liquidation fee
        return [
            *[
                {
                    'name': perp_market_config['name'],
                    'type': 'perpetual',
                    'base_currency': perp_market_config['name'].split('-')[0],
                    'quote_currency': 'USDC',
                    'maker_fees': - (1 / 1e4),
                    'taker_fees': (4 / 1e4)
                }
                for perp_market_config in self.perp_market_configs
            ],
            *[
                {
                    'name': serum3_market_config['name'],
                    'type': 'spot',
                    'base_currency': serum3_market_config['name'].split('/')[0],
                    'quote_currency': serum3_market_config['name'].split('/')[1],
                    'maker_fees': - (0.5 / 1e4),
                    'taker_fees': (1 / 1e4)
                }
                for serum3_market_config in self.serum_market_configs
            ]
        ]

    async def orderbook_l2(self, symbol: str, depth: int = 100):
        # TODO: Validate that the symbol entered is valid

        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        match market_type:
            case 'spot':
                serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

                serum_market_index = serum_market_config['marketIndex']

                serum_market = [
                    serum_market
                    for serum_market in self.serum_markets
                    if serum_market.market_index == serum_market_index
                ][0]

                serum_market_external = [
                    serum_market_external
                    for serum_market_external in self.serum_markets_external
                    if serum_market_external.state.public_key() == serum_market.serum_market_external
                ][0]

                response = await self.connection.get_multiple_accounts([
                    serum_market_external.state.bids(),
                    serum_market_external.state.asks()
                ])

                [raw_bids, raw_asks] = response.value

                [bids, asks] = [
                    OrderBook.from_bytes(serum_market_external.state, raw_bids.data),
                    OrderBook.from_bytes(serum_market_external.state, raw_asks.data)
                ]

                orderbook = {
                    'symbol': symbol,
                    'bids': [[bid.price, bid.size] for bid in bids.get_l2(depth)],
                    'asks': [[ask.price, ask.size] for ask in asks.get_l2(depth)],
                    'slot': response.context.slot
                }

                return orderbook
            case 'perpetual':
                perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

                perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

                accounts = await self.connection.get_multiple_accounts([perp_market.bids, perp_market.asks, perp_market.oracle])

                [raw_bids, raw_asks, raw_oracle] = accounts.value

                [bids, asks] = [BookSide.decode(raw_bids.data), BookSide.decode(raw_asks.data)]

                oracle = pyth.PRICE.parse(raw_oracle.data)

                oracle_price = oracle.agg.price * (Decimal(10) ** oracle.expo)

                orderbook = {
                    'symbol': symbol,
                    'bids': BookSideItems('bids', bids, perp_market, oracle_price).l2()[:depth],
                    'asks': BookSideItems('asks', asks, perp_market, oracle_price).l2()[:depth],
                    'slot': accounts.context.slot
                }

                return orderbook

    async def incremental_orderbook_l2(self, symbol: str, depth: int = 50):
        # TODO: Validate the symbol exists
        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        yield await self.orderbook_l2(symbol, depth)

        match market_type:
            case 'spot':
                serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

                serum_market_index = serum_market_config['marketIndex']

                serum_market = [
                    serum_market
                    for serum_market in self.serum_markets
                    if serum_market.market_index == serum_market_index
                ][0]

                serum_market_external = [
                    serum_market_external
                    for serum_market_external in self.serum_markets_external
                    if serum_market_external.state.public_key() == serum_market.serum_market_external
                ][0]

                orderbook = {
                    'symbol': symbol,
                    'bids': None,
                    'asks': None
                }

                async def snapshots(side):
                    async with connect(self.connection._provider.endpoint_uri.replace('https://', 'wss://')) as websocket:
                        await websocket.account_subscribe(getattr(serum_market_external.state, side)(), Processed, 'jsonParsed')

                        async for message in websocket:
                            for submessage in message:
                                if not isinstance(submessage, AccountNotification):
                                    continue

                                orders = [
                                    [order.price, order.size]
                                    for order in OrderBook.from_bytes(
                                        serum_market_external.state,
                                        submessage.result.value.data
                                    ).get_l2(depth)
                                ]

                                yield side, orders, submessage.result.context.slot

                async with aiostream.stream.merge(*[stream for stream in [snapshots(side) for side in ['bids', 'asks']]]).stream() as streamer:
                    async for side, orders, slot in streamer:
                        orderbook[side] = orders

                        if not all([orderbook['bids'], orderbook['asks']]):
                            continue

                        yield orderbook

            case 'perpetual':
                perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

                perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

                state = {
                    'oracle_price': None
                }

                orderbook = {
                    'symbol': symbol,
                    'bids': None,
                    'asks': None
                }

                async def oracle_price():
                    async with connect(self.connection._provider.endpoint_uri.replace('https://', 'wss://')) as websocket:
                        await websocket.account_subscribe(perp_market.oracle, Processed, 'jsonParsed')

                        async for message in websocket:
                            for submessage in message:
                                if not isinstance(submessage, AccountNotification):
                                    continue

                                oracle = pyth.PRICE.parse(submessage.result.value.data)

                                yield {
                                    'channel': 'oracle_price',
                                    'symbol': symbol,
                                    'value': float(Decimal(str(oracle.agg.price)) * Decimal(10) ** Decimal(str(oracle.expo)))
                                }

                async def book(side):
                    async with connect(self.connection._provider.endpoint_uri.replace('https://', 'wss://')) as websocket:
                        await websocket.account_subscribe(getattr(perp_market, side), Processed, 'jsonParsed')

                        async for message in websocket:
                            for submessage in message:
                                if not isinstance(submessage, AccountNotification):
                                    continue

                            if not state['oracle_price']:
                                continue

                            yield {
                                'channel': 'book',
                                'side': side,
                                'orders': BookSideItems(side, BookSide.decode(submessage.result.value.data), perp_market, state['oracle_price']).l2()
                            }

                async with aiostream.stream.merge(*[oracle_price(), *[book(side) for side in ['bids', 'asks']]]).stream() as streamer:
                    async for message in streamer:
                        match message['channel']:
                            case 'oracle_price':
                                state['oracle_price'] = message['value']
                            case 'book':
                                orderbook[message['side']] = message['orders']

                        if not all([orderbook['bids'], orderbook['asks']]):
                            continue

                        yield orderbook

    async def fills(self, symbol: str):
        # TODO: Validate that the symbol entered is valid

        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        match market_type:
            case 'spot':
                raise NotImplementedError("Spot markets fills retrieval isn't implemented yet")
            case 'perpetual':
                perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

                perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

                response = await self.connection.get_account_info(perp_market.event_queue)

                event_queue = EventQueue.decode(response.value.data)

                fills = sorted(
                    [
                        fill
                        for fill in [
                            FillEvent.layout.parse(bytes([event.event_type] + event.padding))
                            for event in event_queue.buf
                            if event.event_type == 0
                        ]
                    ],
                    key=lambda fill: fill.seq_num
                )

                return {
                    'symbol': symbol,
                    'fills': [
                        {
                            'side': 'bids' if fill.taker_side else 'asks',
                            'price': PerpMarketHelper.price_lots_to_ui(perp_market, fill.price),
                            'size': PerpMarketHelper.base_lots_to_ui(perp_market, fill.quantity),
                            'taker': fill.taker,
                            'maker': fill.maker,
                            'timestamp': fill.timestamp
                        }
                        for fill in fills
                    ],
                    'slot': response.context.slot
                }

    async def incremental_fills(self, symbol: str):
        # TODO: Validate the symbol exists
        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        match market_type:
            case 'spot':
                raise NotImplementedError("Spot markets incremental fills retrieval isn't implemented yet")
            case 'perpetual':
                perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

                perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

                lead = None

                async with connect(self.connection._provider.endpoint_uri.replace('https://', 'wss://')) as websocket:
                    await websocket.account_subscribe(perp_market.event_queue, Processed, 'jsonParsed')

                    async for message in websocket:
                        for submessage in message:
                            if not isinstance(submessage, AccountNotification):
                                continue

                            event_queue = EventQueue.decode(submessage.result.value.data)

                            fills = sorted(
                                [
                                    fill
                                    for fill in [
                                        FillEvent.layout.parse(bytes([event.event_type] + event.padding))
                                        for event in event_queue.buf
                                        if event.event_type == 0
                                    ]
                                    if fill.taker != PublicKey(0)
                                ],
                                key=lambda fill: fill.seq_num
                            )

                            if lead:
                                fills = [fill for fill in fills if fill.seq_num > lead]

                                if len(fills) == 0:
                                    continue

                            yield {
                                'symbol': symbol,
                                'is_snapshot': False,
                                'fills': [
                                    {
                                        'side': 'bids' if fill.taker_side else 'asks',
                                        'price': PerpMarketHelper.price_lots_to_ui(perp_market, fill.price),
                                        'size': PerpMarketHelper.base_lots_to_ui(perp_market, fill.quantity),
                                        'taker': fill.taker,
                                        'maker': fill.maker,
                                        'timestamp': fill.timestamp
                                    }
                                    for fill in fills
                                ],
                                'slot': submessage.result.context.slot
                            }

                            lead = fills[-1:][0].seq_num

    async def funding_rate(self, symbol: str):
        """

        Returns instantaneous funding rate for the day: funding is continuously
        applied on every interaction with a perp position. The rate is further
        multiplied by the time elapsed since it was last applied (capped to max. 1hr).

        :param symbol:
        :return: instantaneous funding rate in % form
        """
        perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

        perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

        accounts = await self.connection.get_multiple_accounts([perp_market.bids, perp_market.asks, perp_market.oracle])

        raw_bids, raw_asks, raw_oracle = accounts.value

        bids, asks, oracle = BookSide.decode(raw_bids.data), BookSide.decode(raw_asks.data), pyth.PRICE.parse(raw_oracle.data)

        oracle_price = float(Decimal(str(oracle.agg.price)) * Decimal(10) ** oracle.expo)

        [bids, asks] = [BookSideItems('bids', bids, perp_market, oracle_price), BookSideItems('asks', asks, perp_market, oracle_price)]

        min_funding, max_funding = float(perp_market.min_funding.to_decimal()), float(perp_market.max_funding.to_decimal())

        impact_quantity = PerpMarketHelper.base_lots_to_ui(perp_market, perp_market.impact_quantity)

        bid, ask = [bids.impact_price(impact_quantity), asks.impact_price(impact_quantity)]

        if bid and ask:
            mid_price = (bid + ask) / 2

            funding = min(max(mid_price / oracle_price - 1, min_funding), max_funding)
        elif bid:
            funding = max_funding
        elif ask:
            funding = min_funding
        else:
            funding = 0

        return funding * 100

    async def get_mango_account(self, public_key: str): return await MangoAccount.fetch(self.connection, PublicKey(public_key))

    async def get_all_mango_accounts(self):
        response = await self.connection.get_program_accounts(
            MANGO_PROGRAM_ID,
            encoding='base64'
        )

        mango_accounts = []

        for entry in response.value: # TODO: Use proper filters instead
            try:
                mango_account = MangoAccount.decode(entry.account.data, entry.pubkey)
            except anchorpy.error.AccountInvalidDiscriminator:
                continue

            mango_accounts.append(mango_account)

        return mango_accounts

    def _health_remaining_accounts(
        self,
        retriever: Literal['fixed', 'scanning'],
        banks: [Bank],
        perp_markets: [PerpMarket],
        mango_account: MangoAccount
    ) -> [AccountMeta]:
        health_remaining_account_pks = []

        match retriever:
            case 'fixed':
                token_indices = [token.token_index for token in mango_account.tokens]

                for bank in banks:
                    if bank.token_index not in token_indices:
                        index = [
                            idx for idx, token in enumerate(mango_account.tokens)
                            if token.token_index == 65535
                            if token_indices[idx] == 65535
                        ][0]

                        token_indices[index] = bank.token_index

                mint_infos = [
                    [mint_info for mint_info in self.mint_infos if mint_info.token_index == token_index][0]
                    for token_index in token_indices if token_index != 65535
                ]

                health_remaining_account_pks.extend([mint_info.banks[0] for mint_info in mint_infos])

                health_remaining_account_pks.extend([mint_info.oracle for mint_info in mint_infos])

                perp_market_indices = [perp.market_index for perp in mango_account.perps]

                for perp_market in perp_markets:
                    if perp_market.perp_market_index not in perp_market_indices:
                        index = [
                            idx for idx, perp in enumerate(mango_account.perps)
                            if perp.market_index == 65535
                            if perp_market_indices[idx] == 65535
                        ][0] = perp_market.perp_market_index

                        perp_market_indices[index] = perp_market.perp_market_index

                perp_markets = [
                    perp_market for perp_market in self.perp_markets
                    if perp_market.perp_market_index in [perp_index for perp_index in perp_market_indices if perp_index != 65535]
                ]

                perp_market_pks = [
                    PublicKey(perp_market_config['publicKey']) for perp_market_config in self.group_config['perpMarkets']
                    if perp_market_config['marketIndex'] in [perp_market.perp_market_index for perp_market in perp_markets]
                ]

                health_remaining_account_pks.extend(perp_market_pks)

                health_remaining_account_pks.extend([perp_market.oracle for perp_market in perp_markets])

                health_remaining_account_pks.extend([serum3.open_orders for serum3 in mango_account.serum3 if serum3.market_index != 65535])
            case 'scanning':
                raise NotImplementedError()

        remaining_accounts: [AccountMeta] = [
            AccountMeta(pubkey=remaining_account_pk, is_writable=False, is_signer=False)
            for remaining_account_pk in health_remaining_account_pks
        ]

        return remaining_accounts

    def make_serum3_create_open_orders_ix(self, mango_account: MangoAccount, symbol: str):
        # TODO: Check whether the symbol matches a valid Serum market name
        serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

        serum_market_index = serum_market_config['marketIndex']

        serum_market = [
            serum_market
            for serum_market in self.serum_markets
            if serum_market.market_index == serum_market_index
        ][0]

        [open_orders_pk, nonce] = PublicKey.find_program_address(
            [
                bytes('Serum3OO', 'utf-8'),
                bytes(mango_account.public_key),
                bytes(PublicKey(serum_market_config['publicKey']))
            ],
            MANGO_PROGRAM_ID
        )

        serum3_create_open_orders_accounts: Serum3CreateOpenOrdersAccounts = {
            'group': mango_account.group,
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'serum_market': PublicKey(serum_market_config['publicKey']),
            'serum_program': serum_market.serum_program,
            'serum_market_external': serum_market.serum_market_external,
            'open_orders': open_orders_pk,
            'payer': mango_account.owner
        }

        serum3_create_open_orders_ix = serum3_create_open_orders(serum3_create_open_orders_accounts)

        return serum3_create_open_orders_ix

    def make_serum3_place_order_ix(self, mango_account: MangoAccount, symbol: str, side: Literal['bids', 'asks'], price: float, size: float):
        serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

        serum_market_index = serum_market_config['marketIndex']

        serum_market = [
            serum_market
            for serum_market in self.serum_markets
            if serum_market.market_index == serum_market_index
        ][0]

        serum_market_external = [
            serum_market_external
            for serum_market_external in self.serum_markets_external
            if serum_market_external.state.public_key() == serum_market.serum_market_external
        ][0]

        limit_price = serum_market_external.state.price_number_to_lots(price)

        max_base_qty = serum_market_external.state.base_size_number_to_lots(size)

        order_type = {'limit': serum3_order_type.Limit(), 'immediate_or_cancel': serum3_order_type.ImmediateOrCancel()}['limit']

        max_native_quote_qty_without_fees = limit_price * max_base_qty

        is_maker = order_type == serum3_order_type.PostOnly()

        fees = {True: - (0.5 / 1e4), False: (1 / 1e4)}[is_maker]

        max_native_quote_qty_including_fees = max_native_quote_qty_without_fees + round(max_native_quote_qty_without_fees * fees)

        client_order_id = round(time.time_ns() / 1e6)

        serum3_place_order_args: Serum3PlaceOrderArgs = {
            'side': {'bids': serum3_side.Bid(), 'asks': serum3_side.Ask()}[side],
            'limit_price': limit_price,
            'max_base_qty': max_base_qty,
            'max_native_quote_qty_including_fees': max_native_quote_qty_including_fees,
            'self_trade_behavior': serum3_self_trade_behavior.DecrementTake(),
            'order_type': order_type,
            'client_order_id': client_order_id,
            'limit': 10
        }

        serum3 = [serum3 for serum3 in mango_account.serum3 if serum3.market_index == serum_market_index][0]

        if serum3 is None:
            raise Exception('serum3 account not found')

        payer_token_index = {
            'bids': serum_market.quote_token_index,
            'asks': serum_market.base_token_index
        }[side]

        bank = [bank for bank in self.banks if bank.token_index == payer_token_index][0]

        banks_config = [
            {
                'tokenIndex': token_config['tokenIndex'],
                'publicKey': PublicKey(token_config['banks'][0]['publicKey'])
            }
            for token_config in self.group_config['tokens']
            if token_config['active']
        ]

        bank_config = [bank_config for bank_config in banks_config if bank_config['tokenIndex'] == bank.token_index][0]

        serum_market_external_vault_signer_address = PublicKey.create_program_address([
            bytes(serum_market.serum_market_external),
            serum_market_external.state.vault_signer_nonce().to_bytes(8, 'little')
        ], SERUM_PROGRAM_ID)

        serum3_place_order_accounts: Serum3PlaceOrderAccounts = {
            'group': mango_account.group,
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'open_orders': serum3.open_orders,
            'serum_market': PublicKey(serum_market_config['publicKey']),
            'serum_program': SERUM_PROGRAM_ID,
            'serum_market_external': serum_market.serum_market_external,
            'market_bids': serum_market_external.state.bids(),
            'market_asks': serum_market_external.state.asks(),
            'market_event_queue': serum_market_external.state.event_queue(),
            'market_request_queue': serum_market_external.state.request_queue(),
            'market_base_vault': serum_market_external.state.base_vault(),
            'market_quote_vault': serum_market_external.state.quote_vault(),
            'market_vault_signer': serum_market_external_vault_signer_address,
            'payer_bank': bank_config['publicKey'],
            'payer_vault': bank.vault,
            'payer_oracle': bank.oracle
        }

        remaining_accounts = self._health_remaining_accounts('fixed', [], [], mango_account)

        serum3_place_order_ix = serum3_place_order(
            serum3_place_order_args,
            serum3_place_order_accounts,
            remaining_accounts=remaining_accounts
        )

        return serum3_place_order_ix

    def make_perp_place_order_ix(self, mango_account: MangoAccount, symbol, side, price, size):
        perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

        perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

        quote_decimals = 6

        def to_native(ui_amount: float, decimals: float) -> int:
            return int(ui_amount * 10 ** decimals)

        def ui_price_to_lots(perp_market: PerpMarket, price: float) -> int:
            return int(to_native(price, quote_decimals) * perp_market.base_lot_size / (perp_market.quote_lot_size * 10 ** perp_market.base_decimals))

        def ui_base_to_lots(perp_market: PerpMarket, size: float) -> int:
            return int(to_native(size, perp_market.base_decimals) // perp_market.base_lot_size)

        perp_place_order_args: PerpPlaceOrderArgs = {
            'side': {'bids': Bid, 'asks': Ask}[side],
            'price_lots': ui_price_to_lots(perp_market, price),
            'max_base_lots': ui_base_to_lots(perp_market, size),
            'max_quote_lots': sys.maxsize,
            'client_order_id': int(time.time() * 1e3),
            'order_type': place_order_type.Limit(),
            'reduce_only': False,
            'expiry_timestamp': 0,
            'limit': 10
        }

        perp_place_order_accounts: PerpPlaceOrderAccounts = {
            'group': perp_market.group,
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'perp_market': PublicKey(perp_market_config['publicKey']),
            'bids': perp_market.bids,
            'asks': perp_market.asks,
            'event_queue': perp_market.event_queue,
            'oracle': perp_market.oracle
        }

        remaining_accounts = self._health_remaining_accounts('fixed', [[bank for bank in self.banks if bank.token_index == 0][0]], [perp_market], mango_account)

        perp_place_order_ix = perp_place_order(
            perp_place_order_args,
            perp_place_order_accounts,
            remaining_accounts=remaining_accounts
        )

        return perp_place_order_ix

    async def place_order(
        self,
        symbol: str,
        side: Literal['bids', 'asks'],
        price: float,
        size: float,
        mango_account: MangoAccount,
        keypair: Keypair
    ):
        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        match market_type:
            case 'spot':
                serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

                serum_market_index = serum_market_config['marketIndex']

                try:
                    serum3 = [serum3 for serum3 in mango_account.serum3 if serum3.market_index == serum_market_index][0]
                except IndexError:
                    logging.error(f"Open orders account for {symbol} not found, creating one...")

                    serum3_create_open_orders_ix = self.make_serum3_create_open_orders_ix(mango_account, symbol)

                    recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

                    tx = Transaction()

                    tx.add(serum3_create_open_orders_ix)

                    response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

                    await self.connection.confirm_transaction(response.value)

                    logging.error(f"Open orders account created for {symbol}: {response}. Please re-launch the program.")

                serum3_place_order_ix = self.make_serum3_place_order_ix(
                    mango_account,
                    symbol,
                    side,
                    price,
                    size
                )

                tx = Transaction()

                recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

                tx.add(serum3_place_order_ix)

                response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

                return response
            case 'perpetual':
                tx = Transaction()

                recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

                perp_place_order_ix = self.make_perp_place_order_ix(mango_account, symbol, side, price, size)

                tx.add(perp_place_order_ix)

                response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

                return response

    def make_serum3_cancel_all_orders_ix(self, mango_account: MangoAccount, symbol: str):
        serum_market_config = [serum3_market_config for serum3_market_config in self.group_config['serum3Markets'] if serum3_market_config['name'] == symbol][0]

        serum_market_index = serum_market_config['marketIndex']

        serum_market = [
            serum_market
            for serum_market in self.serum_markets
            if serum_market.market_index == serum_market_index
        ][0]

        try:
            serum3 = [serum3 for serum3 in mango_account.serum3 if serum3.market_index == serum_market_index][0]
        except IndexError as error:
            print(error)

        serum_market_external = [
            serum_market_external
            for serum_market_external in self.serum_markets_external
            if serum_market_external.state.public_key() == serum_market.serum_market_external
        ][0]

        serum3_cancel_all_orders_args: Serum3CancelAllOrdersArgs = {
            'limit': 10
        }

        serum3_cancel_all_orders_accounts: Serum3CancelAllOrdersAccounts = {
            'group': mango_account.group,
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'open_orders': serum3.open_orders,
            'serum_market': PublicKey(serum_market_config['publicKey']),
            'serum_program': serum_market.serum_program,
            'serum_market_external': serum_market.serum_market_external,
            'market_bids': serum_market_external.state.bids(),
            'market_asks': serum_market_external.state.asks(),
            'market_event_queue': serum_market_external.state.event_queue()
        }

        serum3_cancel_all_orders_ix = serum3_cancel_all_orders(
            serum3_cancel_all_orders_args,
            serum3_cancel_all_orders_accounts
        )

        return serum3_cancel_all_orders_ix

    def make_perp_cancel_all_orders_ix(self, mango_account: MangoAccount, symbol: str):
        perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

        perp_market = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

        perp_cancel_all_orders_args: PerpCancelAllOrdersArgs = {
            'limit': 10
        }

        perp_cancel_all_orders_accounts: PerpCancelAllOrdersAccounts = {
            'group': perp_market.group,
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'perp_market': PublicKey(perp_market_config['publicKey']),
            'bids': perp_market.bids,
            'asks': perp_market.asks
        }

        perp_cancel_all_orders_ix = perp_cancel_all_orders(perp_cancel_all_orders_args, perp_cancel_all_orders_accounts)

        return perp_cancel_all_orders_ix

    async def cancel_all_orders(self, mango_account: MangoAccount, keypair: Keypair, symbol: str):
        market_type = {'PERP': 'perpetual', 'USDC': 'spot'}[re.split(r"[-|/]", symbol)[1]]

        match market_type:
            case 'spot':
                tx = Transaction()

                recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

                serum3_cancel_all_orders_ix = self.make_serum3_cancel_all_orders_ix(mango_account, symbol)

                tx.add(serum3_cancel_all_orders_ix)

                response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

                return response
            case 'perpetual':
                tx = Transaction()

                recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

                tx.recent_blockhash = str(recent_blockhash)

                perp_cancel_all_orders_ix = self.make_perp_cancel_all_orders_ix(mango_account, symbol)

                tx.add(perp_cancel_all_orders_ix)

                response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

                return response

    async def balances(self, mango_account: MangoAccount):
        # TODO: Clean up this mess
        return [
            {
                'symbol': meta['symbol'],
                'balance': float(
                    meta['token_indexed_position'] * (
                        meta['bank_deposit_index'] if meta['token_indexed_position'] > 0
                        else meta['bank_borrow_index']
                    )
                    /
                    meta['bank_mint_decimals']
                )
            }
            for meta in
            [
                {
                    'symbol': token_config['symbol'],
                    'token_indexed_position': Decimal(token.indexed_position.val) / divider,
                    'bank_mint_decimals': 10 ** bank.mint_decimals,
                    'bank_deposit_index': Decimal(bank.deposit_index.val) / divider,
                    'bank_borrow_index': Decimal(bank.borrow_index.val) / divider,
                }
                for token, bank, token_config, divider in
                [
                    [
                        token,
                        [bank for bank in self.banks if bank.token_index == token.token_index][0],
                        [token_config for token_config in self.group_config['tokens'] if token_config['tokenIndex'] == token.token_index][0],
                        Decimal(2 ** (8 * 6))
                    ]
                    for token in mango_account.tokens
                    if token.token_index != 65535
                ]
            ]
        ]

    def make_place_perp_pegged_order_ix(
        self,
        mango_account: MangoAccount,
        symbol: str,
        side: Literal['bids', 'asks'],
        price_offset: float,
        peg_limit: float,
        quantity: float,
        max_quote_quantity: float = None,
        client_order_id: int = int(time.time()),
        expiry_timestamp: int = 0,
        limit: int = 10,
        reduce_only: bool = False
    ):
        perp_market_config = [perp_market_config for perp_market_config in self.group_config['perpMarkets'] if perp_market_config['name'] == symbol][0]

        perp_market: PerpMarket = [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_market_config['marketIndex']][0]

        perp_place_order_pegged_args: PerpPlaceOrderPeggedArgs = {
            'side': {'bids': Bid, 'asks': Ask}[side],
            'price_offset_lots': PerpMarketHelper.ui_price_to_lots(perp_market, price_offset),
            'peg_limit': PerpMarketHelper.ui_price_to_lots(perp_market, peg_limit),
            'max_base_lots': PerpMarketHelper.ui_base_to_lots(perp_market, quantity),
            'max_quote_lots': PerpMarketHelper.ui_quote_to_lots(perp_market, max_quote_quantity) if max_quote_quantity else RUST_I64_MAX,
            'client_order_id': client_order_id,
            'order_type': Limit,
            'reduce_only': reduce_only,
            'expiry_timestamp': expiry_timestamp,
            'limit': limit,
            'max_oracle_staleness_slots': -1
        }

        perp_place_order_pegged_accounts: PerpPlaceOrderPeggedAccounts = {
            'group': PublicKey(self.group_config['publicKey']),
            'account': mango_account.public_key,
            'owner': mango_account.owner,
            'perp_market': PublicKey(perp_market_config['publicKey']),
            'bids': perp_market.bids,
            'asks': perp_market.asks,
            'event_queue': perp_market.event_queue,
            'oracle': perp_market.oracle
        }

        remaining_accounts = self._health_remaining_accounts(
            'fixed',
            [bank for bank in self.banks if bank.token_index == 0],
            [perp_market],
            mango_account
        )

        return perp_place_order_pegged(
            perp_place_order_pegged_args,
            perp_place_order_pegged_accounts,
            remaining_accounts=remaining_accounts
        )

    async def place_perp_pegged_order(
        self,
        mango_account: MangoAccount,
        keypair: Keypair,
        symbol: str,
        side: Literal['bids', 'asks'],
        price_offset: float,
        peg_limit: float,
        quantity: float,
        max_quote_quantity: float = None,
        client_order_id: int = int(time.time()),
        expiry_timestamp: int = 0,
        limit: int = 10,
        reduce_only: bool = False
    ):
        tx = Transaction()

        recent_blockhash = str((await self.connection.get_latest_blockhash()).value.blockhash)

        tx.add(
            self.make_place_perp_pegged_order_ix(
                mango_account,
                symbol,
                side,
                price_offset,
                peg_limit,
                quantity,
                max_quote_quantity,
                client_order_id,
                expiry_timestamp,
                limit,
                reduce_only
            )
        )

        response = await self.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash)

        return response

    async def equity(self, mango_account: MangoAccount):
        oracle_price_by_token_index = {}

        oracle_price_by_oracle_pk = {}

        for bank, raw_oracle in zip(
            self.banks,
            await asyncio.gather(*[self.connection.get_account_info(bank.oracle) for bank in self.banks])
        ):
            name = bytes(bank.name).decode().strip('\x00')

            if name == 'USDC':
                oracle_price = 1
            else:
                # TODO: Handle MNGO's oracle (not Pyth)

                oracle = pyth.PRICE.parse(raw_oracle.value.data)

                oracle_price = oracle.agg.price * (Decimal(10) ** oracle.expo)

            oracle_price_by_token_index[bank.token_index] = oracle_price

            oracle_price_by_oracle_pk[bank.oracle] = oracle_price

        balance_by_token_index = {}

        for token, bank in [
            (
                token,
                [bank for bank in self.banks if bank.token_index == token.token_index][0]
            )
            for token in filter(TokenPositionHelper.is_active, mango_account.tokens)
        ]:
            oracle_price = oracle_price_by_token_index[token.token_index]

            balance_by_token_index[token.token_index] = Decimal(str(TokenPositionHelper.balance(token, bank))) * oracle_price

        active_open_orders = [open_orders for open_orders in mango_account.serum3 if Serum3OrdersHelper.is_active(open_orders)]

        for open_orders, open_orders_external in zip(
            active_open_orders,
            await asyncio.gather(*[
                AsyncOpenOrdersAccount.load(self.connection, str(open_orders.open_orders))
                for open_orders
                in active_open_orders
            ])
        ):
            balance_by_token_index[open_orders.base_token_index] += ((open_orders_external.base_token_total) * oracle_price_by_token_index[open_orders.base_token_index] * Decimal(10 ** (6 - bank.mint_decimals))) / Decimal(1e6)

            balance_by_token_index[open_orders.base_token_index] += (open_orders_external.quote_token_total * oracle_price_by_token_index[open_orders.quote_token_index]) / Decimal(1e6)

        token_equity = sum(balance_by_token_index.values())

        perp_equity = sum([
            PerpPositionHelper.equity(perp_position, perp_market, oracle_price_by_oracle_pk[perp_market.oracle])
            for perp_position, perp_market in [
                (
                    perp_position,
                    [perp_market for perp_market in self.perp_markets if perp_market.perp_market_index == perp_position.market_index][0]
                )
                for perp_position in mango_account.perps
                if PerpPositionHelper.is_active(perp_position)
            ]
        ])

        return token_equity + perp_equity

    async def health_ratio(self, mango_account: MangoAccount, health_type: Literal['init', 'maint', 'liquidation_end']):
        health_type: HealthTypeKind = {
            'init': Init(),
            'maint': Maint(),
            'liquidation_end': LiquidationEnd()
        }[health_type]

        # Build the health cache

        token_positions = [
            token_position
            for token_position in mango_account.tokens
            if token_position.token_index != 65535
        ]

        banks = [
            [bank for bank in self.banks if bank.token_index == token_position.token_index][0]
            for token_position in token_positions
        ]

        raw_oracles = await self.connection.get_multiple_accounts([bank.oracle for bank in banks])

        def oracle_price_from_account_info(account: Account):
            match str(account.owner):
                case 'FsJ3A3u2vn5cTVofAjvy6y5kwABJAqYWpe4975bi2epH': # Pyth
                    oracle = pyth.PRICE.parse(account.data)

                    return oracle.agg.price * (Decimal(10) ** oracle.expo)
                case '4MangoMjqJ2firMokCjjGgoK8d4MXcrgL7XJaL3w6fVg': # For now it's always USDC
                    return Decimal(1)

        oracle_prices = [
            oracle_price_from_account_info(account)
            for account in raw_oracles.value
        ]

        token_infos = [
            TokenInfo(
                bank.token_index,
                bank.maint_asset_weight,
                bank.init_asset_weight,
                BankHelper.scaled_init_asset_weight(bank, PricesHelper.liab(prices, Init())),
                bank.maint_liab_weight,
                bank.init_liab_weight,
                BankHelper.scaled_init_liab_weight(bank, PricesHelper.liab(prices, Init())),
                prices,
                I80F48.from_decimal(TokenPositionHelper.balance(token_position, bank))
            )
            for bank, token_position, prices
            in zip(
                banks,
                token_positions,
                [
                    Prices(
                        oracle=I80F48.from_decimal(oracle_price * Decimal(10 ** (6 - bank.mint_decimals))),
                        stable=I80F48.from_decimal(Decimal(bank.stable_price_model.stable_price))
                    )
                    for bank, oracle_price in zip(banks, oracle_prices)
                ]
            )
        ]

        serum3_infos = []

        for open_orders, open_orders_external in zip(
                MangoAccountHelper.active_serum3_orders(mango_account),
                await asyncio.gather(*[
                    AsyncOpenOrdersAccount.load(self.connection, str(open_orders.open_orders))
                    for open_orders
                    in MangoAccountHelper.active_serum3_orders(mango_account)
                ])
        ):
            base_index, base_info = [
                (index, token_info)
                for index, token_info in enumerate(token_infos)
                if token_info.token_index == open_orders.base_token_index
            ][0]

            if not base_info:
                raise ValueError(f"Base token info not found for market index {open_orders.market_index}")

            quote_index, quote_info = [
                (index, token_info)
                for index, token_info in enumerate(token_infos)
                if token_info.token_index == open_orders.quote_token_index
            ][0]

            if not quote_info:
                raise ValueError(f"Quote token info not found for market index {open_orders.market_index}")

            reserved_base = open_orders_external.base_token_total - open_orders_external.base_token_free

            reserved_quote = open_orders_external.quote_token_total - open_orders_external.quote_token_free

            serum3_infos.append(
                Serum3Info(
                    reserved_base=I80F48.from_decimal(Decimal(reserved_base)),
                    reserved_quote=I80F48.from_decimal(Decimal(reserved_quote)),
                    base_index=base_index,
                    quote_index=quote_index,
                    market_index=open_orders.market_index,
                    has_zero_funds=False
                )
            )

        perp_positions = MangoAccountHelper.active_perp_positions(mango_account)

        perp_markets = [
            [
                perp_market
                for perp_market in self.perp_markets
                if perp_market.perp_market_index == perp_position.market_index
            ][0]
            for perp_position in perp_positions

        ]

        perp_market_oracle_prices = [
            oracle_price_from_account_info(raw_oracle)
            for raw_oracle in (await self.connection.get_multiple_accounts([perp_market.oracle for perp_market in perp_markets])).value
        ]

        perp_infos = []

        for perp_position, perp_market, perp_market_oracle_price in zip(
            perp_positions,
            perp_markets,
            perp_market_oracle_prices
        ):
            base_lots = perp_position.base_position_lots + perp_position.taker_base_lots

            unsettled_funding = PerpPositionHelper.unsettled_funding(perp_position, perp_market)

            taker_quote = perp_position.taker_quote_lots * perp_market.quote_lot_size

            quote_current = I80F48.from_decimal(perp_position.quote_position_native.to_decimal() - unsettled_funding + Decimal(taker_quote))

            perp_info = PerpInfo(
                perp_market.perp_market_index,
                perp_market.maint_base_asset_weight,
                perp_market.init_base_asset_weight,
                perp_market.maint_base_liab_weight,
                perp_market.init_base_liab_weight,
                perp_market.maint_overall_asset_weight,
                perp_market.init_overall_asset_weight,
                perp_market.base_lot_size,
                base_lots,
                perp_position.bids_base_lots,
                perp_position.asks_base_lots,
                quote_current,
                Prices(
                    oracle=I80F48.from_decimal(perp_market_oracle_price * Decimal(10 ** (6 - perp_market.base_decimals))),
                    stable=I80F48.from_decimal(Decimal(perp_market.stable_price_model.stable_price))
                ),
                PerpPositionHelper.has_open_orders(perp_position),
                PerpPositionHelper.has_open_fills(perp_position)
            )

            perp_infos.append(perp_info)

        health_cache = HealthCache(
            token_infos,
            serum3_infos,
            perp_infos,
            False
        )

        # Compute assets and liabilities

        assets = 0

        liabs = 0

        for token_info in health_cache.token_infos:
            contrib = TokenInfoHelper.health_contribution(token_info, health_type)

            if contrib > 0:
                assets += contrib
            else:
                liabs -= contrib

        def get_serum3_reservations(health_type: HealthTypeKind):
            token_max_reserved = [0 for _ in range(0, len(token_infos))]

            serum3_reserved: [Serum3Reserved] = []

            for serum3_info in serum3_infos:
                quote, base = token_infos[serum3_info.quote_index], token_infos[serum3_info.base_index]

                reserved_base, reserved_quote = serum3_info.reserved_base.to_decimal(), serum3_info.reserved_quote.to_decimal()

                quote_asset = PricesHelper.asset(quote.prices, health_type)

                base_liab = PricesHelper.liab(base.prices, health_type)

                all_reserved_as_base = reserved_base + reserved_quote * quote_asset / base_liab

                base_asset = PricesHelper.asset(base.prices, health_type)

                quote_liab = PricesHelper.liab(quote.prices, health_type)

                all_reserved_as_quote = reserved_quote + reserved_base * base_asset / quote_liab

                token_max_reserved[serum3_info.base_index] += all_reserved_as_base

                token_max_reserved[serum3_info.quote_index] += all_reserved_as_quote

                serum3_reserved.append(Serum3Reserved(all_reserved_as_base, all_reserved_as_quote))

            return {
                'token_max_reserved': token_max_reserved,
                'serum3_reserved': serum3_reserved
            }

        res = get_serum3_reservations(Maint())

        for index, serum3_info in enumerate(serum3_infos):
            contrib = Serum3InfoHelper.health_contribution(
                serum3_info,
                health_type,
                health_cache.token_infos,
                res['token_max_reserved'],
                res['serum3_reserved'][index]
            ).to_decimal()

            if contrib > 0:
                assets += contrib
            else:
                liabs -= contrib

        for perp_info in health_cache.perp_infos:
            contrib = PerpInfoHelper.health_contribution(perp_info, health_type).to_decimal()

            if contrib > 0:
                assets += contrib
            else:
                liabs -= contrib

        if liabs > 0.001:
            return 100 * (assets - liabs) / liabs
        else:
            return sys.maxsize


import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class GroupJSON(typing.TypedDict):
    creator: str
    group_num: int
    admin: str
    fast_listing_admin: str
    mngo_token_index: int
    padding: list[int]
    insurance_vault: str
    insurance_mint: str
    bump: int
    testing: int
    version: int
    buyback_fees: int
    buyback_fees_mngo_bonus_factor: float
    address_lookup_tables: list[str]
    security_admin: str
    deposit_limit_quote: int
    ix_gate: int
    buyback_fees_swap_mango_account: str
    buyback_fees_expiry_interval: int
    reserved: list[int]


@dataclass
class Group:
    discriminator: typing.ClassVar = b"\xd1\xf9\xd0?\xb6Y\xba\xfe"
    layout: typing.ClassVar = borsh.CStruct(
        "creator" / BorshPubkey,
        "group_num" / borsh.U32,
        "admin" / BorshPubkey,
        "fast_listing_admin" / BorshPubkey,
        "mngo_token_index" / borsh.U16,
        "padding" / borsh.U8[2],
        "insurance_vault" / BorshPubkey,
        "insurance_mint" / BorshPubkey,
        "bump" / borsh.U8,
        "testing" / borsh.U8,
        "version" / borsh.U8,
        "buyback_fees" / borsh.U8,
        "buyback_fees_mngo_bonus_factor" / borsh.F32,
        "address_lookup_tables" / BorshPubkey[20],
        "security_admin" / BorshPubkey,
        "deposit_limit_quote" / borsh.U64,
        "ix_gate" / borsh.U128,
        "buyback_fees_swap_mango_account" / BorshPubkey,
        "buyback_fees_expiry_interval" / borsh.U64,
        "reserved" / borsh.U8[1824],
    )
    creator: PublicKey
    group_num: int
    admin: PublicKey
    fast_listing_admin: PublicKey
    mngo_token_index: int
    padding: list[int]
    insurance_vault: PublicKey
    insurance_mint: PublicKey
    bump: int
    testing: int
    version: int
    buyback_fees: int
    buyback_fees_mngo_bonus_factor: float
    address_lookup_tables: list[PublicKey]
    security_admin: PublicKey
    deposit_limit_quote: int
    ix_gate: int
    buyback_fees_swap_mango_account: PublicKey
    buyback_fees_expiry_interval: int
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["Group"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Group"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Group"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Group":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Group.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            creator=dec.creator,
            group_num=dec.group_num,
            admin=dec.admin,
            fast_listing_admin=dec.fast_listing_admin,
            mngo_token_index=dec.mngo_token_index,
            padding=dec.padding,
            insurance_vault=dec.insurance_vault,
            insurance_mint=dec.insurance_mint,
            bump=dec.bump,
            testing=dec.testing,
            version=dec.version,
            buyback_fees=dec.buyback_fees,
            buyback_fees_mngo_bonus_factor=dec.buyback_fees_mngo_bonus_factor,
            address_lookup_tables=dec.address_lookup_tables,
            security_admin=dec.security_admin,
            deposit_limit_quote=dec.deposit_limit_quote,
            ix_gate=dec.ix_gate,
            buyback_fees_swap_mango_account=dec.buyback_fees_swap_mango_account,
            buyback_fees_expiry_interval=dec.buyback_fees_expiry_interval,
            reserved=dec.reserved,
        )

    def to_json(self) -> GroupJSON:
        return {
            "creator": str(self.creator),
            "group_num": self.group_num,
            "admin": str(self.admin),
            "fast_listing_admin": str(self.fast_listing_admin),
            "mngo_token_index": self.mngo_token_index,
            "padding": self.padding,
            "insurance_vault": str(self.insurance_vault),
            "insurance_mint": str(self.insurance_mint),
            "bump": self.bump,
            "testing": self.testing,
            "version": self.version,
            "buyback_fees": self.buyback_fees,
            "buyback_fees_mngo_bonus_factor": self.buyback_fees_mngo_bonus_factor,
            "address_lookup_tables": list(
                map(lambda item: str(item), self.address_lookup_tables)
            ),
            "security_admin": str(self.security_admin),
            "deposit_limit_quote": self.deposit_limit_quote,
            "ix_gate": self.ix_gate,
            "buyback_fees_swap_mango_account": str(
                self.buyback_fees_swap_mango_account
            ),
            "buyback_fees_expiry_interval": self.buyback_fees_expiry_interval,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: GroupJSON) -> "Group":
        return cls(
            creator=PublicKey(obj["creator"]),
            group_num=obj["group_num"],
            admin=PublicKey(obj["admin"]),
            fast_listing_admin=PublicKey(obj["fast_listing_admin"]),
            mngo_token_index=obj["mngo_token_index"],
            padding=obj["padding"],
            insurance_vault=PublicKey(obj["insurance_vault"]),
            insurance_mint=PublicKey(obj["insurance_mint"]),
            bump=obj["bump"],
            testing=obj["testing"],
            version=obj["version"],
            buyback_fees=obj["buyback_fees"],
            buyback_fees_mngo_bonus_factor=obj["buyback_fees_mngo_bonus_factor"],
            address_lookup_tables=list(
                map(lambda item: PublicKey(item), obj["address_lookup_tables"])
            ),
            security_admin=PublicKey(obj["security_admin"]),
            deposit_limit_quote=obj["deposit_limit_quote"],
            ix_gate=obj["ix_gate"],
            buyback_fees_swap_mango_account=PublicKey(
                obj["buyback_fees_swap_mango_account"]
            ),
            buyback_fees_expiry_interval=obj["buyback_fees_expiry_interval"],
            reserved=obj["reserved"],
        )
